"""The ``settings`` subcommand action."""

from __future__ import annotations

from dataclasses import asdict
from functools import partial
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.configuration_subsystem import PresentableSettingsEntries
from ansible_navigator.configuration_subsystem import PresentableSettingsEntry
from ansible_navigator.configuration_subsystem import to_effective
from ansible_navigator.configuration_subsystem import to_presentable
from ansible_navigator.configuration_subsystem import to_sample
from ansible_navigator.configuration_subsystem import to_schema
from ansible_navigator.configuration_subsystem import to_sources
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.configuration_subsystem.definitions import Constants
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.steps import StepType
from ansible_navigator.steps import TypedStep
from ansible_navigator.ui_framework import Color
from ansible_navigator.ui_framework import CursesLine
from ansible_navigator.ui_framework import CursesLinePart
from ansible_navigator.ui_framework import CursesLines
from ansible_navigator.ui_framework import Decoration
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.utils.print import print_to_stdout

from . import _actions as actions
from . import run_action


if TYPE_CHECKING:
    from ansible_navigator.app_public import AppPublic


def color_menu(colno: int, colname: str, entry: PresentableSettingsEntry) -> tuple[int, int]:
    """Color the menu.

    Args:
        colno: Column number
        colname: Column name
        entry: Column value

    Returns:
        Constants that curses uses to color a line of text
    """
    if entry.default:
        return Color.GREEN, Color.BLACK
    return Color.YELLOW, Color.BLACK


CONTENT_HEADING_DEFAULT = "{name} (current value/default value: {current_value})"
CONTENT_HEADING_NOT_DEFAULT = (
    "{name} (current value: {current_value})  (default value: {default_value})"
)


def content_heading(obj: PresentableSettingsEntry, screen_w: int) -> CursesLines:
    """Create a heading for the setting entry showing.

    Args:
        obj: The content going to be shown
        screen_w: The current screen width

    Returns:
        The heading
    """
    if obj.default:
        text = CONTENT_HEADING_DEFAULT.format(**asdict(obj))
        color = Color.GREEN
    else:
        text = CONTENT_HEADING_NOT_DEFAULT.format(**asdict(obj))
        color = Color.YELLOW

    fill_characters = screen_w - len(text) + 1
    heading_line = f"{text}{' ' * fill_characters}"

    line_part = CursesLinePart(
        column=0,
        string=heading_line,
        color=color,
        decoration=Decoration.UNDERLINE,
    )
    return CursesLines((CursesLine((line_part,)),))


@actions.register
class Action(ActionBase):
    """The action class for the settings subcommand."""

    KEGEX = r"^se(?:ttings)?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:settings`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="settings")
        self._settings: PresentableSettingsEntries

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Handle the ``settings`` subcommand in mode ``interactive``.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("settings requested")
        self._prepare_to_run(app, interaction)
        self._settings = to_presentable(self._args)
        self.steps.append(self._build_main_menu())

        while True:
            self._calling_app.update()
            self._take_step()

            if not self.steps:
                break

            if self.steps.current.name == "quit":
                return self.steps.current

        self._prepare_to_exit(interaction)
        return None

    def run_stdout(self) -> RunStdoutReturn:
        """Handle settings in mode stdout.

        Returns:
            RunStdoutReturn
        """
        self._logger.debug("settings requested in stdout mode")

        dump = partial(
            print_to_stdout,
            use_color=self._args.display_color,
        )
        if self._args.entry("settings_schema").value.source is not Constants.DEFAULT_CFG:
            dump(content=to_schema(self._args), content_format=ContentFormat.JSON)
            return RunStdoutReturn(message="", return_code=0)

        dumped = False
        if self._args.settings_effective:
            dump(content=to_effective(self._args), content_format=ContentFormat.YAML)
            dumped = True
        if self._args.settings_sample:
            dump(content=to_sample(self._args)[0], content_format=ContentFormat.YAML_TXT)
            dumped = True
        if self._args.settings_sources:
            dump(content=to_sources(self._args), content_format=ContentFormat.YAML)
            dumped = True
        if dumped:
            return RunStdoutReturn(message="", return_code=0)

        dump(content=to_presentable(self._args), content_format=ContentFormat.YAML)
        return RunStdoutReturn(message="", return_code=0)

    def _build_main_menu(self) -> TypedStep[Any]:
        """Build the main menu of settings.

        Returns:
            The settings menu definition
        """
        step = TypedStep[PresentableSettingsEntry](
            name="all_options",
            columns=["name", "default", "source", "current"],
            select_func=self._build_settings_content,
            step_type=StepType.MENU,
        )
        step.value = self._settings
        return step

    def _build_settings_content(self) -> TypedStep[Any]:
        """Build the content for one settings entry.

        Returns:
            The option's content
        """
        step = TypedStep[PresentableSettingsEntry](
            name="setting_content",
            step_type=StepType.CONTENT,
        )
        step.index = self.steps.current.index
        step.value = self._settings
        return step

    def _take_step(self) -> None:
        """Take one step in the stack of steps."""
        result = None
        if isinstance(self.steps.current, Interaction):
            result = run_action(self.steps.current.name, self.app, self.steps.current)
        elif isinstance(self.steps.current, TypedStep):
            if self.steps.current.show_func:
                current_index = self.steps.current.index
                self.steps.current.show_func()
                self.steps.current.index = current_index

            if self.steps.current.step_type is StepType.MENU:
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    columns=self.steps.current.columns,
                    color_menu_item=color_menu,
                )
            elif self.steps.current.step_type is StepType.CONTENT:
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    index=self.steps.current.index,
                    content_heading=content_heading,
                )

        if result is None:
            self.steps.back_one()
        else:
            self.steps.append(result)



"""Build a menu."""

from __future__ import annotations

import curses
import enum
import re

from typing import TYPE_CHECKING
from typing import Any

from .curses_defs import CursesLine
from .curses_defs import CursesLinePart
from .curses_defs import CursesLines
from .utils import convert_percentage
from .utils import distribute


if TYPE_CHECKING:
    from collections.abc import Callable

    from ansible_navigator.content_defs import ContentBase
    from ansible_navigator.content_defs import ContentTypeSequence

    from .ui_config import UIConfig


class MenuBuilder:
    """Build a menu from list of dicts."""

    def __init__(
        self,
        progress_bar_width: int,
        screen_width: int,
        number_colors: int,
        color_menu_item: Callable[..., Any],
        ui_config: UIConfig,
    ) -> None:
        """Initialize the menu builder.

        Args:
            progress_bar_width: The width of the progress bar
            screen_width: The current screen width
            number_colors: The number of colors the current terminal
                supports
            color_menu_item: The callback for adding color to menu
                entries
            ui_config: The current user interface configuration
        """
        self._number_colors = number_colors
        self._progress_bar_width = progress_bar_width
        self._screen_width = screen_width
        self._color_menu_item = color_menu_item
        self._ui_config = ui_config

    def build(
        self,
        dicts: ContentTypeSequence,
        cols: list[str],
        indices: tuple[int, ...],
    ) -> tuple[CursesLines, CursesLines]:
        """Build menu main entry point.

        Args:
            cols: he columns (keys) to use in the dicts
            dicts: A list of dicts
            indices: A range of what's showing in the UI

        Returns:
            The heading and body of a menu
        """
        return self._menu(dicts, cols, indices)

    def _menu(
        self,
        dicts: ContentTypeSequence,
        cols: list[str],
        indices: tuple[int, ...],
    ) -> tuple[CursesLines, CursesLines]:
        """Build a text menu from a list of dicts given columns(root keys).

        Args:
            dicts: A list of dicts
            cols: The columns (keys) to use in the dicts
            indices: A range of what's showing in the UI

        Returns:
            the heading and body of the menu
        """
        line_prefix_w = len(str(len(dicts))) + len("|")

        for idx in indices:
            convert_percentage(dicts[idx], cols, self._progress_bar_width)

        lines = [[str(dicts[idx].get(c)) for c in cols] for idx in indices]
        column_widths = [
            max(len(str(v)) for v in c)
            for c in zip(*lines + [[re.sub("^__", "", col) for col in cols]], strict=False)
        ]
        # add a space
        column_widths = [c + 1 for c in column_widths]

        available = self._screen_width - line_prefix_w - 1  # scrollbar width
        adjusted_column_widths = distribute(available, column_widths)

        col_starts = [0]
        for idx, column_width in enumerate(adjusted_column_widths):
            col_starts.append(column_width + col_starts[idx])

        menu_layout = (col_starts, cols, adjusted_column_widths)
        header = self._menu_header_line(menu_layout)

        menu_layout = (col_starts, cols, adjusted_column_widths, header)  # type: ignore
        menu_lines = self._menu_lines(dicts, menu_layout, indices)
        return CursesLines((header,)), menu_lines

    def _menu_header_line(self, menu_layout: Any) -> CursesLine:
        """Generate the menu header line.

        Args:
            menu_layout: A tuple of menu details:

                * ``menu_layout[0]``: ``List[int]``, the starting in for each column
                * ``menu_layout[1]``: ``List[str]``, the columns of the menu
                * ``menu_layout[2]``: ``List[int]``, the adjusted column widths

        Returns:
            The menu header line
        """
        _column_starts, cols, _adjusted_column_widths = menu_layout
        line_parts = tuple(
            self._menu_header_line_part(colno, menu_layout) for colno in range(len(cols))
        )
        return CursesLine(line_parts)

    @staticmethod
    def _menu_header_line_part(colno: int, menu_layout: tuple[list[Any], ...]) -> CursesLinePart:
        """Generate one part of the menu header line.

        Args:
            colno: The column number
            menu_layout: A tuple of menu details:

                * ``menu_layout[0]``: ``List[int]``, the starting in for each column
                * ``menu_layout[1]``: ``List[str]``, the columns of the menu
                * ``menu_layout[2]``: ``List[int]``, the adjusted column widths

        Returns:
            The menu head line
        """
        col_starts, cols, adjusted_column_widths = menu_layout
        coltext = re.sub("^__", "", cols[colno])
        coltext = coltext.replace("_", " ")
        adj_entry = coltext[0 : adjusted_column_widths[colno]].capitalize()
        # right justify header if progress
        if cols[colno] == "__progress":
            return CursesLinePart(
                column=col_starts[colno] + adjusted_column_widths[colno] - len(adj_entry),
                string=adj_entry,
                color=0,
                decoration=curses.A_UNDERLINE,
            )
        return CursesLinePart(
            column=col_starts[colno],
            string=adj_entry,
            color=0,
            decoration=curses.A_UNDERLINE,
        )

    def _menu_lines(
        self,
        dicts: ContentTypeSequence,
        menu_layout: Any,
        indices: tuple[int, ...],
    ) -> CursesLines:
        """Generate all the menu lines.

        Args:
            dicts: A list of dicts from which the menu will be generated
            indices: A range of what's showing in the UI
            menu_layout: A tuple of menu details:

            * ``menu_layout[0]``: ``List[int]``, the starting in for each column
            * ``menu_layout[1]``: ``List[str]``, the columns of the menu
            * ``menu_layout[2]``: ``List[int]``, the adjusted column widths
            * ``menu_layout[3]``: ``CursesLine``, the menu header, used to determine justification

        Returns:
            The menu lines
        """
        return CursesLines(tuple(self._menu_line(dicts[idx], menu_layout) for idx in indices))

    def _menu_line(
        self,
        menu_entry: dict[str, Any] | ContentBase[Any],
        menu_layout: tuple[list[Any], ...],
    ) -> CursesLine:
        """Generate one the menu line.

        Args:
            menu_entry: One dict from which the menu line will be generated
            menu_layout: A tuple of menu details:

            * ``menu_layout[0]``: ``List[int]``, the starting in for each column
            * ``menu_layout[1]``: ``List[str]``, the columns of the menu
            * ``menu_layout[2]``: ``List[int]``, the adjusted column widths
            * ``menu_layout[3]``: ``CursesLine``, the menu header, used to determine justification

        Returns:
            A menu line
        """
        _column_starts, cols, _adjusted_column_widths, _header = menu_layout
        menu_line = (menu_entry.get(c) for c in cols)
        line_parts = (
            self._menu_line_part(colno, coltext, menu_entry, menu_layout)
            for colno, coltext in enumerate(menu_line)
        )
        return CursesLine(tuple(line_parts))

    def _menu_line_part(
        self,
        colno: int,
        coltext: Any,
        menu_entry: dict[str, Any] | ContentBase[Any],
        menu_layout: tuple[list[Any], ...],
    ) -> CursesLinePart:
        # pylint: disable=too-many-locals
        """Generate one menu line part.

        Args:
            colno: The column number of the line part
            coltext: The text to be placed at the given column
            menu_entry: The dict from which the menu line will be
                generated
            menu_layout: A tuple of menu details:

            * ``menu_layout[0]``: ``List[int]``, the starting in for each column
            * ``menu_layout[1]``: ``List[str]``, the columns of the menu
            * ``menu_layout[2]``: ``List[int]``, the adjusted column widths
            * ``menu_layout[3]``: ``CursesLine``, the menu header, used to determine justification

        Returns:
            A menu line part
        """
        column_starts, cols, adjusted_column_widths, header = menu_layout

        color, decoration = self._color_menu_item(colno, cols[colno], menu_entry)

        text = str(coltext)[0 : adjusted_column_widths[colno]]

        is_bool = isinstance(coltext, bool)
        is_number = isinstance(coltext, int | float)
        is_enum = isinstance(coltext, enum.Enum)
        is_duration = cols[colno].lower() == "__duration"
        is_progress = cols[colno].lower() == "__progress"

        if is_bool:
            # booleans are left justified
            print_at = column_starts[colno]
        elif (is_number and not is_enum) or is_duration:
            # right justify on header if int, float or "duration"
            print_at = column_starts[colno] + len(header[colno][1]) - len(text)
        elif is_progress:
            # right justify in column if progress indicator
            print_at = column_starts[colno] + adjusted_column_widths[colno] - len(text)
        else:
            # left justify
            print_at = column_starts[colno]
        return CursesLinePart(column=print_at, string=str(text), color=color, decoration=decoration)



"""Utilities for working with dictionaries using paths."""

from __future__ import annotations

import copy
import operator

from enum import Enum
from functools import reduce
from typing import TYPE_CHECKING
from typing import Any


if TYPE_CHECKING:
    from collections.abc import MutableMapping


class MergeBehaviors(Enum):
    """The merge behaviors."""

    LIST_LIST_EXTEND = "append list to list"
    LIST_LIST_REPLACE = "replace list with list"
    LIST_APPEND = "append to list"
    LIST_REPLACE = "replace list"
    LIST_SORT = "sort resulting list"
    LIST_UNIQUE = "only unique values in resulting list"
    DICT_DICT_UPDATE = "update left dict with right dict"
    DICT_DICT_REPLACE = "replace left dict with right dict"


def get_with_path(content: MutableMapping[Any, Any], path: str) -> Any:
    """Get a value from a path in a dictionary.

    Args:
        content: The content of the settings file
        path: The path to the value

    Returns:
        The value at the path
    """
    return reduce(operator.getitem, path.split("."), content)


def check_path(content: MutableMapping[Any, Any], path: str) -> bool:
    """Check if a path exists in a dictionary.

    Args:
        content: The content of the settings file
        path: The path to the value

    Returns:
        Whether the path exists
    """
    try:
        get_with_path(content, path)
    except (KeyError, TypeError):
        return False
    return True


def delete_with_path(content: MutableMapping[Any, Any], path: str) -> None:
    """Delete a value from a path in a dictionary.

    Args:
        content: The content of the settings file
        path: The path to the value
    """
    parts = path.split(".")
    del reduce(operator.getitem, parts[:-1], content)[parts[-1]]


def ascendants_from_path(path: str) -> list[str]:
    """Get the ascendants of a path.

    Args:
        path: The path to the value

    Returns:
        The ascendants of the path
    """
    parts = path.split(".")
    return [path.rsplit(".", i)[0] for i in range(len(parts))]


def descendants_to_path(path: str) -> list[str]:
    """Get the descendants to a path.

    Args:
        path: The path to the value

    Returns:
        The descendants to the path
    """
    parts = path.split(".")
    return [path.rsplit(".", i)[0] for i in reversed(range(len(parts)))]


def remove_and_delete_empty_ascendants(content: MutableMapping[Any, Any], path: str) -> None:
    """Remove and delete empty ascendants.

    Args:
        content: The content of the settings file
        path: The path to the value
    """
    ascendants = ascendants_from_path(path)
    delete_with_path(content, ascendants.pop(0))
    while ascendants:
        ascendant = ascendants.pop(0)
        branch_value = get_with_path(content, ascendant)
        if branch_value == {}:
            delete_with_path(content, ascendant)
        else:
            break


def _merge_into_list(
    behaviors: tuple[MergeBehaviors, ...],
    nested: dict[Any, Any],
    part: str,
    value: bool | list[Any] | float | str | dict[Any, Any],
) -> None:
    """Merge a value into an existing list at the given key.

    Args:
        behaviors: The merge behaviors.
        nested: The parent dict containing the list.
        part: The key of the list in nested.
        value: The value to merge.

    Raises:
        ValueError: If no matching behavior is specified.
    """
    if isinstance(value, list):
        if MergeBehaviors.LIST_LIST_EXTEND in behaviors:
            nested[part].extend(value)
        elif MergeBehaviors.LIST_LIST_REPLACE in behaviors:
            nested[part] = value
        else:
            msg = "No behavior specified for LIST_LIST"
            raise ValueError(msg)
    elif MergeBehaviors.LIST_APPEND in behaviors:
        nested[part].append(value)
    elif MergeBehaviors.LIST_REPLACE in behaviors:
        nested[part] = value
        return
    else:
        msg = "No behavior specified for LIST_*"
        raise ValueError(msg)

    if MergeBehaviors.LIST_UNIQUE in behaviors:
        nested[part] = list(dict.fromkeys(nested[part]))
    if MergeBehaviors.LIST_SORT in behaviors:
        nested[part].sort()


def _merge_into_dict(
    behaviors: tuple[MergeBehaviors, ...],
    nested: dict[Any, Any],
    part: str,
    value: dict[Any, Any],
) -> None:
    """Merge a dict value into an existing dict at the given key.

    Args:
        behaviors: The merge behaviors.
        nested: The parent dict containing the target dict.
        part: The key of the dict in nested.
        value: The dict value to merge.

    Raises:
        ValueError: If no matching behavior is specified.
    """
    if MergeBehaviors.DICT_DICT_UPDATE in behaviors:
        nested[part].update(value)
    elif MergeBehaviors.DICT_DICT_REPLACE in behaviors:
        nested[part] = value
    else:
        msg = "No behavior specified for DICT_DICT"
        raise ValueError(msg)


def _place_at_leaf(
    behaviors: tuple[MergeBehaviors, ...],
    nested: dict[Any, Any],
    part: str,
    value: bool | list[Any] | float | str | dict[Any, Any],
) -> None:
    """Place a value at a leaf key in the nested dict.

    Args:
        behaviors: The merge behaviors.
        nested: The parent dict.
        part: The leaf key.
        value: The value to place.
    """
    if isinstance(nested.get(part), list):
        _merge_into_list(behaviors, nested, part, value)
    elif isinstance(nested.get(part), dict) and isinstance(value, dict):
        _merge_into_dict(behaviors, nested, part, value)
    else:
        nested[part] = value


def place_at_path(
    behaviors: tuple[MergeBehaviors, ...],
    content: dict[Any, Any],
    path: str,
    value: bool | list[Any] | float | str | dict[Any, Any],
) -> dict[Any, Any]:
    """Place a value at a path in a dictionary.

    Args:
        behaviors: The merge behaviors
        content: The content of the settings file
        path: The path to the value
        value: The value to place

    Raises:
        ValueError: If something can't be done

    Returns:
        The updated content
    """
    if (
        MergeBehaviors.DICT_DICT_REPLACE in behaviors
        and MergeBehaviors.DICT_DICT_UPDATE in behaviors
    ):
        msg = "Can't use both DICT_DICT_REPLACE and DICT_DICT_UPDATE behaviors"
        raise ValueError(msg)
    if (
        MergeBehaviors.LIST_LIST_EXTEND in behaviors
        and MergeBehaviors.LIST_LIST_REPLACE in behaviors
    ):
        msg = "Can't use both LIST_LIST_EXTEND and LIST_LIST_REPLACE behaviors"
        raise ValueError(msg)

    copied_content = copy.deepcopy(content)
    nested = copied_content
    if path in ("", None):
        if isinstance(value, dict):
            if MergeBehaviors.DICT_DICT_REPLACE in behaviors:
                return value
            if MergeBehaviors.DICT_DICT_UPDATE in behaviors:
                return {**nested, **value}
        msg = "Cannot place non dict at root of dict"
        raise ValueError(msg)

    parts = path.split(".")
    last_idx = len(parts) - 1
    for idx, part in enumerate(parts):
        if idx == last_idx:
            _place_at_leaf(behaviors, nested, part, value)
        else:
            if part not in nested:
                nested[part] = {}
            nested = nested[part]
    return copied_content


def move_to_path(
    behaviors: tuple[MergeBehaviors, ...],
    content: dict[Any, Any],
    new_path: str,
    old_path: str,
) -> dict[Any, Any]:
    """Move a value to a path in a dictionary.

    Args:
        behaviors: The merge behaviors
        content: The content of the settings file
        old_path: The path to the value
        new_path: The path to the value

    Returns:
        The updated content
    """
    copied_content = copy.deepcopy(content)
    if new_path == old_path:
        return copied_content

    value = get_with_path(content=copied_content, path=old_path)
    delete_with_path(content=copied_content, path=old_path)
    updated_content = place_at_path(
        content=copied_content,
        path=new_path,
        value=value,
        behaviors=behaviors,
    )
    return updated_content



"""Step abstractions for actions."""

from __future__ import annotations

from collections import deque
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import TypeVar


if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Sequence


class Step:
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    """One step in the flow of things.

    The ``TypedStep`` below should be used for all new actions.
    """

    def __init__(
        self,
        name: str,
        step_type: str,
        value: list[dict[str, str]] | Any,
        columns: list[str] | None = None,
        index: int | None = None,
        select_func: Callable[[], Step] | None = None,
        show_func: Callable[[], None] | None = None,
    ) -> None:
        """Initialize the instance of a step.

        Args:
            name: The name of the step
            step_type: The type of step
            value: The data corresponding to the step
            columns: The columns to show when the data is presented as a
                menu
            index: The currently selected entry in the data
            select_func: The function to be called when the user selects
                an entry
            show_func: The function to call prior to an entry being
                shown to the user
        """
        self._index = index
        self._index_changed = False
        self._value = value or []
        self._value_changed = False
        self.columns = columns or []
        self.name = name
        self.next = None
        self.previous = None
        self.select_func = select_func
        self.show_func = show_func
        self.type = step_type

    @property
    def changed(self) -> bool:
        """Return if this has changed.

        Returns:
            Boolean if index or value has changed
        """
        return self._index_changed or self._value_changed

    @changed.setter
    def changed(self, value: bool) -> None:
        """Set the changed value.

        Args:
            value (bool): The value to set
        """
        self._value_check(value, bool)
        self._index_changed = value
        self._value_changed = value

    @property
    def index(self) -> int | None:
        """Return the index.

        Returns:
            Index (should be ``int``)
        """
        return self._index

    @index.setter
    def index(self, index: int | None) -> None:
        """Set the index.

        Args:
            index (int): The index
        """
        self._value_check(index, (int, type(None)))
        self._index_changed = self._index != index
        self._index = index

    @property
    def selected(self) -> dict[str, Any] | None:
        """Return the selected item.

        Returns:
            The selected item.
        """
        if self._index is None or not self._value:
            return None
        return self._value[self._index % len(self._value)]

    @property
    def value(self) -> list[dict[str, Any]]:
        """Return the value.

        Returns:
            The value
        """
        return self._value

    @value.setter
    def value(self, value: list[dict[str, Any]]) -> None:
        """Set the value and changed is needed.

        Args:
            value (list): List of dicts
        """
        self._value_check(value, list)
        self._value_changed = self._value != value
        self._value = value

    @staticmethod
    def _value_check(value: Any, want: type | tuple[type, ...]) -> None:
        """Check some expected type against a value's type.

        Args:
            value: Some value for comparison
            want: The desired type for value

        Raises:
            ValueError: If value type doesn't match wanted type
        """
        if not isinstance(value, want):
            msg = f"wanted {want}, got {type(value)}"
            raise TypeError(msg)


class StepType(Enum):
    """Type of step, either menu or content."""

    MENU = "menu"
    CONTENT = "content"


T = TypeVar("T")


@dataclass
class TypedStep(Generic[T]):
    # pylint: disable=too-many-instance-attributes
    """One step in the flow of things.

    This should be used for new actions following the pattern of a
    ``TypedDict`` for the user interface as seen in settings.
    """

    name: str
    step_type: StepType
    _index_changed: bool = False
    _index: int | None = None
    _value_changed: bool = False
    _value: Sequence[T] = field(default_factory=list)
    columns: list[str] | None = None
    select_func: Callable[[], TypedStep[Any]] | None = None
    show_func: Callable[[], None] | None = None

    @property
    def changed(self) -> bool:
        """Return the changed flag.

        Returns:
            Indication of change
        """
        return self._index_changed or self._value_changed

    @changed.setter
    def changed(self, value: bool) -> None:
        """Set the changed value.

        Args:
            value: Indication of change
        """
        self._index_changed = value
        self._value_changed = value

    @property
    def index(self) -> int | None:
        """Return the index.

        Returns:
            The index
        """
        return self._index

    @index.setter
    def index(self, index: int | None) -> None:
        """Set the index.

        Args:
            index: The index
        """
        self._index_changed = self.index != index
        self._index = index

    @property
    def selected(self) -> T | None:
        """Return the selected item.

        Returns:
            The selected item
        """
        if self._index is None or not self._value:
            return None
        return self._value[self._index % len(self._value)]

    @property
    def value(self) -> Sequence[T]:
        """Return the value.

        Returns:
            The value
        """
        return self._value

    @value.setter
    def value(self, value: Sequence[T]) -> None:
        """Set the value and value changed if needed.

        Args:
            value: The value for this instance
        """
        self._value_changed = self._value != value
        self._value = value


class Steps(deque[Any]):
    """A custom deque."""

    def back_one(self) -> Any:
        """Pop one step off the dequeue to move backwards.

        Returns:
            The dequeue after pop() or None if no steps are left
        """
        if self:
            return self.pop()
        return None

    @property
    def current(self) -> Any:
        """Return the current step.

        Returns:
            The TypedStep (or Step) currently accessed
        """
        return self[-1]

    @property
    def previous(self) -> Any:
        """Return the previous step.

        Returns:
            The TypedStep (or Step) previously accessed
        """
        return self[-2]



"""Utilities related to the configuration subsystem."""

from __future__ import annotations

import logging

from collections.abc import Iterable
from configparser import ConfigParser
from configparser import ParsingError
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any

from ansible_navigator.command_runner import Command
from ansible_navigator.command_runner import CommandRunner
from ansible_navigator.utils.definitions import ExitMessage
from ansible_navigator.utils.definitions import LogMessage

from .definitions import Constants
from .definitions import SettingsFileType


def create_settings_file_sample(
    settings_path: str,
    placeholder: bool | int | str | dict[Any, Any] | list[Any] = "",
) -> SettingsFileType:
    """Generate a settings file sample.

    Args:
        settings_path: The dot delimited settings file path for a
            settings entry
        placeholder: String used to identify the placement of a settings
            value

    Returns:
        A sample of the settings file
    """
    if "." not in settings_path:
        return SettingsFileType({settings_path: placeholder})
    key, remainder = settings_path.split(".", 1)
    return SettingsFileType({key: create_settings_file_sample(remainder, placeholder)})


def ansible_verison_parser(command: Command) -> None:
    """Parse the output of the ansible command.

    Args:
        command: The result of running the command
    """
    if command.return_code:
        return

    messages: list[LogMessage] = []

    details = {}
    # First line contains the version
    version = command.stdout_lines[0].split(" ", 1)[1].strip()[1:-1]
    details["version"] = version
    # Remaining lines are k = v pairs
    for entry in command.stdout_lines[1:]:
        parts = entry.split("=")
        if len(parts) == 2:
            details[parts[0].strip()] = parts[1].strip()
        else:
            msg = "ansible --version entry: '{entry}' could not be parsed"
            messages.append(LogMessage(level=logging.ERROR, message=msg))

    for key, value in details.items():
        msg = f"ansible --version: '{key}:{value}'"
        messages.append(LogMessage(level=logging.ERROR, message=msg))

    command.details = [details]
    command.messages = messages


@dataclass
class AnsibleConfiguration:
    """Data structure for an ansible.cfg file.

    Attributes:
        contents: The parsed contents of the file.
        text: The text from the file.
        path: The path to the file.
    """

    contents: Constants | dict[str, dict[str, bool | int | float | str]] = Constants.NOT_SET
    text: Constants | list[str] = Constants.NOT_SET
    path: Constants | Path = Constants.NOT_SET


@dataclass
class ParseAnsibleCfgResponse:
    """Data structure for the response of parse_ansible_cfg.

    Attributes:
        messages: Log messages.
        exit_messages: Exit messages.
        config: An ansible configuration.
    """

    messages: list[LogMessage]
    exit_messages: list[ExitMessage]
    config: AnsibleConfiguration = field(default_factory=AnsibleConfiguration)


def parse_ansible_cfg(ee_enabled: bool) -> ParseAnsibleCfgResponse:
    """Find the ansible.cfg file and parse it.

    If running without an EE, use ansible to get it.
    If running with an EE only look in the CWD.
    In the future user-specified volume mounts might be considered as locations.

    Args:
        ee_enabled: Indicates if EE support is enabled

    Returns:
        The ansible.cfg contents
    """
    response = ParseAnsibleCfgResponse(messages=[], exit_messages=[])
    response.config.path = Constants.NONE
    response.config.text = Constants.NONE
    response.config.contents = Constants.NONE

    if ee_enabled:
        msg = "EE support enabled: using current working directory for 'ansible.cfg'"
        response.messages.append(LogMessage(level=logging.DEBUG, message=msg))
        cfg_path = Path.cwd() / "ansible.cfg"
        if not cfg_path.exists():
            msg = (
                "EE support enabled: no 'ansible.cfg' found in current working directory."
                f" Tried {cfg_path!s}"
            )
            response.messages.append(LogMessage(level=logging.INFO, message=msg))
            return response

        msg = "EE support enabled: found 'ansible.cfg' in current working directory."
        response.messages.append(LogMessage(level=logging.INFO, message=msg))

    else:
        msg = "EE support disabled: using 'ansible --version' for 'ansible.cfg'"
        response.messages.append(LogMessage(level=logging.DEBUG, message=msg))
        new_messages, new_exit_messages, version_details = parse_ansible_verison()
        response.messages.extend(new_messages)
        response.exit_messages.extend(new_exit_messages)
        if response.exit_messages or version_details is None:
            return response

        cfg_file_name = version_details.get("config file")
        if cfg_file_name is None:
            msg = "EE support disabled: no 'config file' in 'ansible --version'"
            response.messages.append(LogMessage(level=logging.DEBUG, message=msg))
            return response
        if cfg_file_name == "None":
            msg = "EE support disabled: 'ansible --version' reports no config file"
            response.messages.append(LogMessage(level=logging.INFO, message=msg))
            return response
        cfg_path = Path(cfg_file_name)
        if not cfg_path.exists():
            msg = f"EE support disabled: {cfg_path!s} does not exist"
            response.messages.append(LogMessage(level=logging.DEBUG, message=msg))
            return response

    parser = ConfigParser()
    try:
        parser.read(str(cfg_path))
    except ParsingError as exc:
        exit_msg = f"The ansible configuration file '{cfg_path!s}' could not be parsed."
        response.exit_messages.append(ExitMessage(message=exit_msg))
        response.exit_messages.append(ExitMessage(message=str(exc)))
        return response

    response.config.contents = parser.__dict__.get("_sections", {})
    response.config.path = cfg_path
    response.config.text = cfg_path.read_text(encoding="utf-8").splitlines()
    return response


def parse_ansible_verison(
    path: Path | None = None,
) -> tuple[list[LogMessage], list[ExitMessage], dict[str, Any] | None]:
    """Parse the output of the ansible --version command.

    Args:
        path: The path to the ansible executable

    Returns:
        Log messages, exit messages, and the stdout as a dictionary

    Raises:
        TypeError: if there is a typing error
    """
    messages: list[LogMessage] = []
    exit_messages: list[ExitMessage] = []

    ansible_cmd = "ansible" if path is None else str(path / "ansible")

    command = Command(
        identity="ansible_version",
        command=f"{ansible_cmd} --version",
        post_process=ansible_verison_parser,
    )

    CommandRunner.run_single_process(commands=[command])
    if command.return_code:
        msg = "'ansible --version' reported the following errors:"
        exit_messages.append(ExitMessage(message=msg))
        exit_messages.append(ExitMessage(message=command.stderr))
        return messages, exit_messages, None

    msg = f"ansible --version stdout: '{command.stdout}'"
    messages.append(LogMessage(level=logging.DEBUG, message=msg))
    if not isinstance(command.details, Iterable):
        raise TypeError
    return (
        messages,
        exit_messages,
        command.details[0] if isinstance(command.details, list) else None,
    )



"""Replay subcommand implementation.

A stub, the logic is used from ``:run``, only here as a place
to put the kegex.
"""

from . import _actions as actions
from .run import Action as BaseAction


@actions.register
class Action(BaseAction):
    """``:replay`` command implementation."""

    KEGEX = r"""(?x)
            ^
            (?P<replay>rep(?:lay)?
            (\s(?P<params_replay>\S+))?)
            $"""



# cspell:ignore getpid, gmtime, msecs
"""Navigator entry point."""

from __future__ import annotations

import filecmp
import logging
import os
import shutil
import signal
import sys

from copy import deepcopy
from curses import wrapper
from importlib.metadata import version
from importlib.util import find_spec
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from .action_defs import ActionReturn
from .action_defs import RunInteractiveReturn
from .action_defs import RunReturn
from .action_defs import RunStdoutReturn
from .action_runner import ActionRunner
from .actions import run_action_stdout
from .configuration_subsystem import Constants
from .configuration_subsystem import NavigatorConfiguration
from .image_manager import ImagePuller
from .initialization import error_and_exit_early
from .initialization import parse_and_update
from .logger import setup_logger
from .utils.compatibility import importlib_metadata
from .utils.definitions import ExitMessage
from .utils.definitions import ExitPrefix
from .utils.definitions import LogMessage
from .utils.functions import clear_screen
from .utils.functions import generate_cache_path
from .utils.packaged_data import path_to_file


__version__: Constants | str
try:
    from ._version import version as __version__
except ImportError:
    __version__ = Constants.NOT_SET

if TYPE_CHECKING:
    from .configuration_subsystem.definitions import ApplicationConfiguration


APP_NAME = "ansible-navigator"
PKG_NAME = "ansible_navigator"

logger = logging.getLogger(PKG_NAME)


def cache_scripts() -> None:
    """Cache the scripts used to introspect the container image."""
    scripts = ("catalog_collections.py", "image_introspect.py", "python_latest.sh")
    cache_path = generate_cache_path(app_name=APP_NAME)
    cache_path.mkdir(parents=True, exist_ok=True)
    for script in scripts:
        src = path_to_file(filename=script)
        dst = cache_path / script
        message = f"No update required for {src} to {dst}"
        try:
            if not filecmp.cmp(src, dst):
                shutil.copy(src, dst)
                message = f"Updated {src} to {dst} (outdated)"
        except FileNotFoundError:
            shutil.copy(src, dst)
            message = f"Copied {src} to {dst} (missing)"
        logger.log(level=logging.DEBUG, msg=message)


def log_dependencies() -> list[LogMessage]:
    """Retrieve installed packages and log as debug.

    Returns:
        All packages, version and location
    """
    pkgs = []
    found = []
    messages: list[LogMessage] = []
    for pkg_names in importlib_metadata.packages_distributions().values():
        for pkg_name in pkg_names:
            if pkg_name not in found:
                found.append(pkg_name)
                try:
                    spec = find_spec(pkg_name)
                except ModuleNotFoundError:
                    message = f"Package '{pkg_name}' is missing"
                    messages.append(LogMessage(level=logging.DEBUG, message=message))
                    continue
                _location = spec.origin if spec else ""
                _version = version(pkg_name)
                pkgs.append(f"{pkg_name}=={_version} {_location}")

    pkgs.sort()
    messages = [LogMessage(level=logging.DEBUG, message=pkg) for pkg in pkgs]
    return messages


def pull_image(args: Any) -> None:
    """Pull the image if required.

    Args:
        args: Copy of NavigatorConfiguration
    """
    image_puller = ImagePuller(
        container_engine=args.container_engine,
        image=args.execution_environment_image,
        arguments=args.pull_arguments,
        pull_policy=args.pull_policy,
    )
    image_puller.assess()
    if image_puller.assessment.exit_messages:
        error_and_exit_early(image_puller.assessment.exit_messages)
    if image_puller.assessment.pull_required:
        image_puller.prologue_stdout()
        image_puller.pull_stdout()
    if image_puller.assessment.exit_messages:
        error_and_exit_early(image_puller.assessment.exit_messages)


def run(args: ApplicationConfiguration) -> ActionReturn:
    """Run the appropriate subcommand.

    Args:
        args: The current application settings

    Returns:
        A message to display and a return code
    """
    if args.mode == "stdout":
        try:
            result = run_action_stdout(args.app.replace("-", "_"), args)
        except KeyboardInterrupt:
            logger.warning("Dirty exit, killing the pid")
            os.kill(os.getpid(), signal.SIGTERM)
            return RunStdoutReturn(message="", return_code=1)
        return result
    if args.mode == "interactive":
        try:
            clear_screen()
            wrapper(ActionRunner(args=args).run)
            return RunInteractiveReturn(message="", return_code=0)
        except KeyboardInterrupt:
            logger.warning("Dirty exit, killing the pid")
            os.kill(os.getpid(), signal.SIGTERM)
            return RunInteractiveReturn(message="", return_code=1)
    return RunReturn(message="", return_code=0)


def main() -> None:
    """Start application here."""
    messages: list[LogMessage] = log_dependencies()
    exit_messages: list[ExitMessage] = []

    args = deepcopy(NavigatorConfiguration)
    args.application_version = __version__
    args.internals.initializing = True
    messages.extend(args.internals.initialization_messages)
    exit_messages.extend(args.internals.initialization_exit_messages)

    # may have exit messages e.g., share directory
    # from instantiation of NavigatorConfiguration
    if not exit_messages:
        new_messages, new_exit_messages = parse_and_update(sys.argv[1:], args=args)
        messages.extend(new_messages)
        exit_messages.extend(new_exit_messages)

    # In case of errors, the configuration will have rolled back
    # but a viable log file is still needed, set to default since
    # it cannot be determined if the error is log file location related
    if exit_messages:
        args.entry("log_file").value.current = args.entry("log_file").value.default
        args.entry("log_level").value.current = "debug"
        exit_msg = f"Configuration failed, using default log file location. ({args.log_file})"
        exit_msg += f" Log level set to {args.log_level}"
        exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.NOTE))
        exit_msg = "Review the hints and log file to see what went wrong."
        exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))

    try:
        Path(args.log_file).touch()
        setup_logger(args)
    except Exception as exc:  # noqa: BLE001
        exit_msg = "The log file path or logging engine could not be setup."
        exit_msg += " No log file will be available, please check the log file"
        exit_msg += f" path setting. The error was {exc!s}"
        exit_messages.append(ExitMessage(message=exit_msg))
        error_and_exit_early(exit_messages=exit_messages)

    for log_message in messages:
        logger.log(level=log_message.level, msg=log_message.message)

    if exit_messages:
        for entry in exit_messages:
            logger.log(level=entry.level, msg=entry.message)
        error_and_exit_early(exit_messages=exit_messages)

    os.environ.setdefault("ESCDELAY", "25")

    if args.execution_environment:
        pull_image(args)
    cache_scripts()

    run_return = run(args)
    run_message = f"{run_return.message}\n"
    if run_return.return_code != 0 and run_return.message:
        sys.stderr.write(run_message)
        sys.exit(run_return.return_code)
    elif run_return.return_code != 0:
        sys.exit(run_return.return_code)
    elif run_return.message:
        sys.stdout.write(run_message)


if __name__ == "__main__":
    main()

__all__ = ("NavigatorConfiguration",)



"""Definitions for the command runner."""

from __future__ import annotations

import logging
import multiprocessing
import queue
import subprocess

from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Any


if TYPE_CHECKING:
    from collections.abc import Callable
    from queue import Queue

    from ansible_navigator.utils.definitions import LogMessage

logger = logging.getLogger(__name__)

_WORKER_TIMEOUT = 30


PROCESSES = (multiprocessing.cpu_count() - 1) or 1


@dataclass(frozen=False)
class Command:
    """Data structure for details of a command to be run.

    A ``Command`` is updated after instantiated with details from either
    ``stdout`` or ``stderr``.
    """

    # pylint: disable=too-many-instance-attributes
    identity: str
    command: str
    post_process: Callable[..., Any]
    return_code: int = 0
    stdout: str = ""
    stderr: str = ""
    details: list[Any] | dict[Any, Any] | object = field(default_factory=list)
    errors: str = ""
    messages: list[LogMessage] = field(default_factory=list)

    @property
    def stderr_lines(self) -> list[str]:
        """Produce a list of stderr lines.

        Returns:
            A list of stderr lines
        """
        return self.stderr.splitlines()

    @property
    def stdout_lines(self) -> list[str]:
        """Produce a list of stdout lines.

        Returns:
            A list of stdout lines
        """
        return self.stdout.splitlines()


def run_command(command: Command) -> None:
    """Run a command.

    Args:
        command: Command to be run
    """
    try:
        proc_out = subprocess.run(
            command.command,
            capture_output=True,
            check=True,
            text=True,
            shell=True,
        )
        command.return_code = proc_out.returncode
        command.stdout = proc_out.stdout
    except subprocess.CalledProcessError as exc:
        command.return_code = exc.returncode
        command.stdout = str(exc.stdout)
        command.stderr = str(exc.stderr)


def worker(
    pending_queue: multiprocessing.Queue[Any],
    completed_queue: multiprocessing.Queue[Any],
) -> None:
    """Read pending, run, post process, and place in completed.

    Args:
        pending_queue: All pending commands
        completed_queue: All completed commands
    """
    while True:
        command = pending_queue.get()
        if command is None:
            break
        run_command(command)
        try:
            command.post_process(command)
        except Exception as exc:
            command.errors = str(exc)
            logger.exception("post_process failed for command %s", command.identity)
        completed_queue.put(command)


class CommandRunner:
    """Functionality for running commands."""

    def __init__(self) -> None:
        """Initialize the command runner."""
        self._completed_queue: Queue[Any] | None = None
        self._pending_queue: Queue[Command | None] | None = None

    @staticmethod
    def run_single_process(commands: list[Command]) -> list[Command]:
        """Run commands with a single process.

        Args:
            commands: All commands to be run

        Returns:
            The results from running all commands
        """
        results: list[Command] = []
        for command in commands:
            run_command(command)
            command.post_process(command)
            results.append(command)
        return results

    def run_multi_process(self, commands: list[Command]) -> list[Command]:
        """Run commands with multiple processes.

        Workers are started to read from pending queue.
        Exit when the number of results is equal to the number
        of commands needing to be run.

        Args:
            commands: All commands to be run

        Returns:
            The results from running all commands

        Raises:
            RuntimeError: If all workers terminate before all commands complete
        """
        if self._completed_queue is None:
            self._completed_queue = multiprocessing.Manager().Queue()
        if self._pending_queue is None:
            self._pending_queue = multiprocessing.Manager().Queue()

        workers = self.start_workers(commands)
        results: list[Command] = []
        while len(results) != len(commands):
            try:
                result = self._completed_queue.get(timeout=_WORKER_TIMEOUT)
                results.append(result)
            except queue.Empty:
                alive_workers = [w for w in workers if w.is_alive()]
                if not alive_workers:
                    failed = len(commands) - len(results)
                    msg = (
                        f"All worker processes have terminated but {failed} of "
                        f"{len(commands)} commands did not complete. "
                        f"Worker processes may have crashed."
                    )
                    raise RuntimeError(msg)  # noqa: B904
                logger.debug(
                    "Waiting for %d workers (%d/%d results collected)",
                    len(alive_workers),
                    len(results),
                    len(commands),
                )
        for w in workers:
            w.join()
        return results

    def start_workers(self, jobs: list[Command]) -> list[multiprocessing.Process]:
        """Start the workers.

        Args:
            jobs: List of commands to be run

        Returns:
            The list of started worker processes

        Raises:
            RuntimeError: if the pending queue is not initialized
        """
        worker_count = min(len(jobs), PROCESSES)
        processes: list[multiprocessing.Process] = []
        for _proc in range(worker_count):
            proc = multiprocessing.Process(
                target=worker,
                args=(self._pending_queue, self._completed_queue),
            )
            processes.append(proc)
            proc.start()
        if not self._pending_queue:
            raise RuntimeError
        for job in jobs:
            self._pending_queue.put(job)
        for _proc in range(worker_count):
            self._pending_queue.put(None)
        return processes



"""Base class for apps (actions)."""

from __future__ import annotations

import logging

from copy import deepcopy
from typing import TYPE_CHECKING

from ansible_navigator.actions import kegexes

from .action_defs import RunStdoutReturn
from .app_public import AppPublic
from .configuration_subsystem import Constants as C
from .initialization import parse_and_update
from .steps import Steps
from .ui_framework import Interaction
from .ui_framework import ui
from .ui_framework import warning_notification
from .ui_framework.form_utils import settings_notification
from .utils.definitions import ExitMessage
from .utils.definitions import ExitPrefix
from .utils.definitions import LogMessage


if TYPE_CHECKING:
    from re import Pattern

    from .configuration_subsystem.definitions import ApplicationConfiguration


class ActionBase:
    # pylint: disable=too-many-instance-attributes
    """Base class for actions."""

    def __init__(
        self,
        args: ApplicationConfiguration,
        name: str,
        logger_name: str = __name__,
    ) -> None:
        """Initialize the App class.

        Args:
            args: The current application configuration
            name: The name of the action inheriting this
            logger_name: The name for the logger
        """
        self._logger = logging.getLogger(logger_name)

        self._args: ApplicationConfiguration = self._copy_args(args)
        self._calling_app: AppPublic
        self._interaction: Interaction
        self._name = name
        self._previous_filter: Pattern[str] | None
        self._previous_scroll: int
        self.stdout: list[str] = []
        self.steps = Steps()

    @staticmethod
    def _action_match(entry: str) -> tuple[str, ui.Action] | tuple[None, None]:
        """Attempt to match the user input against the regex provided by each action.

        Args:
            entry: the user input

        Returns:
            The name and matching action or not
        """
        for kegex in kegexes():
            match = kegex.kegex.match(entry)
            if match:
                return kegex.name, ui.Action(match=match, value=entry)
        return None, None

    @property
    def app(self) -> AppPublic:
        """Limit the scope of what is carried between actions.

        This will be passed to other actions to limit the scope of
        what can be mutated internally.

        Returns:
            An instance of AppPublic for the current instance of the
            action

        Raises:
            AttributeError: If the args have not been initialized
        """
        if self._args:
            return AppPublic(
                args=self._args,
                name=self._name,
                rerun=self.rerun,
                stdout=self.stdout,
                steps=self.steps,
                update=self.update,
                write_artifact=self.write_artifact,
            )
        msg = "app passed without args initialized"
        raise AttributeError(msg)

    def no_interactive_mode(self, interaction: Interaction, app: AppPublic) -> None:
        """Show a warning notification that the user interactive mode is not supported.

        Args:
            interaction: The interaction from the user
            app: The app instance
        """
        warning = warning_notification(
            messages=[
                f"The '{self._name}' subcommand is not available while using interactive mode.",
                (
                    "[HINT] Start an additional instance of ansible-navigator"
                    " in a new terminal with mode 'stdout'."
                ),
                f"      e.g. 'ansible-navigator {self._name} --mode stdout",
            ],
        )
        interaction.ui.show_form(warning)

    @staticmethod
    def _copy_args(args: ApplicationConfiguration) -> ApplicationConfiguration:
        """Deepcopy the args.

        Note: Unmount the collection doc cache (CDC) first
        the CDC will get mounted if the child needs it
        in parse and update

        Args:
            args: the current application configuration

        Returns:
            A copy of the current application configuration
        """
        args.internals.collection_doc_cache = C.NOT_SET
        return deepcopy(args)

    def _prepare_to_run(self, app: AppPublic, interaction: Interaction) -> None:
        """Prepare for action run.

        An action can call this to prior to running. This
        will set the scroll to zero and store the state
        of the UI so it can be restored later.

        Args:
            app: The instance of the action
            interaction: The current interaction from the UI
        """
        self._calling_app = app
        self._interaction = interaction
        self._interaction.ui.scroll(0)
        self._previous_scroll = interaction.ui.scroll()
        self._previous_filter = interaction.ui.menu_filter()

    def _prepare_to_exit(self, interaction: Interaction) -> None:
        """Prepare for action exit.

        An action can call this after running to clean up and
        restore the state of the UI prior to the action being
        invoked.

        Args:
            interaction: The current interaction from the UI
        """
        interaction.ui.scroll(self._previous_scroll)
        interaction.ui.menu_filter(self._previous_filter)
        interaction.ui.scroll(0)

    def rerun(self) -> None:
        """Rerun the action.

        Defined in the child class if necessary.
        """

    def run_stdout(self) -> RunStdoutReturn:
        """Provide a message saying subcommand does not support mode stdout.

        Returns:
            Message suggesting mode interactive, return code of 1
        """
        messages = []
        message = f"Subcommand '{self._name}' does not support mode 'stdout'."
        messages.append(ExitMessage(message=message))
        message = "Try again with '--mode interactive'"
        messages.append(ExitMessage(message=message, prefix=ExitPrefix.HINT))
        message = "\n".join(str(message) for message in messages)
        return RunStdoutReturn(message=message, return_code=1)

    def update(self) -> None:
        """Update the action.

        Defined in child if necessary
        """

    def _update_args(
        self,
        params: list[str],
        apply_previous_cli_entries: C = C.ALL,
        attach_cdc: bool = False,
    ) -> bool:
        """Update the current args.

        Pass the params through the configuration subsystem
        log messages and exit_messages as warnings since most will result in a form
        while the exit_messages would have cause a sys.exit(1) from the CLI
        each action should handle them in a manner that does not exit the TUI

        Args:
            params: a sys.argv like list of parameters
            apply_previous_cli_entries: Should previous params from the
                CLI be applied
            attach_cdc: Should the collection doc cache be attached to
                the args.internals

        Returns:
            Indication if the args update succeeded or failed
        """
        messages: list[LogMessage]
        exit_messages: list[ExitMessage]

        messages, exit_messages = parse_and_update(
            params=params,
            apply_previous_cli_entries=apply_previous_cli_entries,
            args=self._args,
            attach_cdc=attach_cdc,
        )

        for message in messages:
            self._logger.log(level=message.level, msg=message.message)

        for exit_msg in exit_messages:
            if exit_msg.level == logging.ERROR:
                self._logger.warning(msg=exit_msg.message)
            else:
                self._logger.log(level=exit_msg.level, msg=exit_msg.message)

        if exit_messages:
            warning = settings_notification(color=self._args.display_color, messages=exit_messages)
            self._interaction.ui.show_form(warning)
            return False
        return True

    def write_artifact(self, filename: str) -> None:
        """Write an artifact file.

        This will likely only be used by the run action.
        Defined in child if necessary

        Args:
            filename: The filename to write to
        """



"""A text input field."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any

from .form_handler_text import FormHandlerText
from .sentinels import Unknown
from .sentinels import nonexistent
from .sentinels import unknown
from .validators import FieldValidators


if TYPE_CHECKING:
    from collections.abc import Callable

    from .curses_window import Window


@dataclass
class FieldText:
    # pylint: disable=too-many-instance-attributes
    """A text input field."""

    name: str
    prompt: str
    current_error: str = ""
    default: Any = nonexistent
    window_handler = FormHandlerText
    response: str | Unknown = unknown
    valid: bool | Unknown = unknown
    validator: Callable[..., Any] = FieldValidators.none
    value: Any = unknown
    win: Window | None = None

    @property
    def formatted_default(self) -> str:
        """Return the default value.

        Returns:
            Default value or empty string
        """
        if self.default is nonexistent:
            return ""
        return f" ({self.default})"

    @property
    def full_prompt(self) -> str:
        """Return the full prompt with the default.

        Returns:
            Prompt with default text
        """
        return self.prompt + self.formatted_default

    def pre_populate(self, value: str) -> None:
        """Pre-populate this text input with a value.

        This is different from a default
        in that it will populate the text input field.

        Args:
            value: Item to populate the input field
        """
        self.conditional_validation(str(value))

    def validate(self, response: str) -> None:
        """Validate the response.

        Args:
            response: Text input response
        """
        if response == "" and self.default is not nonexistent:
            response = str(self.default)

        self.response = response
        validation = self.validator(text=response)
        self.current_error = validation.error_msg
        if validation.error_msg:
            self.valid = False
        else:
            self.valid = True
            self.value = validation.value

    def conditional_validation(self, response: str) -> None:
        """Conditional validation used for tab.

        Only accept the value if it validates, otherwise move along.

        Args:
            response: Text input response
        """
        if response == "" and self.default is not nonexistent:
            response = str(self.default)
        # no response or validator is none
        if response or getattr(self.validator, "__name__", "").endswith("none"):
            validation = self.validator(text=response)
            if not validation.error_msg:
                self.validate(response)
                return
            self.response = response
        else:
            self.response = unknown
        self.valid = unknown
        self.current_error = ""



"""``:log`` command implementation."""

from pathlib import Path

from ansible_navigator.action_base import ActionBase
from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.ui_framework import Interaction

from . import _actions as actions


@actions.register
class Action(ActionBase):
    """``:log`` command implementation."""

    KEGEX = r"^l(?:og)?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:log`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="log")

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction:
        """Execute the ``:log`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction`
        """
        self._logger.debug("log requested")
        self._prepare_to_run(app, interaction)

        auto_scroll = True
        while True:
            self._calling_app.update()
            with Path(self._args.log_file).open(encoding="utf-8") as fh:
                log_contents = fh.read()

            new_scroll = len(log_contents.splitlines())
            if auto_scroll:
                interaction.ui.scroll(new_scroll)

            interaction = interaction.ui.show(obj=log_contents, content_format=ContentFormat.LOG)
            if interaction.name != "refresh":
                break

            if interaction.ui.scroll() < new_scroll and auto_scroll:
                self._logger.debug("auto_scroll disabled")
                auto_scroll = False
            elif interaction.ui.scroll() >= new_scroll and not auto_scroll:
                self._logger.debug("auto_scroll enabled")
                auto_scroll = True

        self._prepare_to_exit(interaction)
        return interaction



"""Helper functions for the ``actions`` package."""

from __future__ import annotations

import functools
import importlib
import importlib.resources as importlib_resources
import logging
import os
import re

from typing import TYPE_CHECKING
from typing import Any
from typing import NamedTuple

from ansible_navigator.ui_framework import error_notification


if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Generator

    from ansible_navigator.action_defs import RunStdoutReturn


logger = logging.getLogger(__name__)


# Basic structure for storing information about one action
class ActionT(NamedTuple):
    """NamedTuple representing an action."""

    name: str
    cls: str
    kegex: str


class Kegex(NamedTuple):
    """NamedTuple representing a kegex."""

    name: str
    kegex: str


# Dictionary with information about all registered actions
_ACTIONS: dict[str, dict[str, Any]] = {}


def _import(package: str, action: str) -> None:
    """Import the given action from a package.

    Args:
        package: The name of the package
        action: The action to import
    """
    importlib.import_module(f"{package}.{action}")


def _import_all(package: str) -> None:
    """Import all actions in a package.

    Args:
        package: The name of the package
    """
    for entry in importlib_resources.files(package).iterdir():
        if entry.is_file() and entry.name.endswith(".py") and not entry.name.startswith("_"):
            _import(package, entry.name[0:-3])


def register(cls: Any) -> Any:
    """Register an action, used as a decorator.

    Args:
        cls: The class to register

    Returns:
        The class after registration
    """
    package, _, action = cls.__module__.rpartition(".")
    pkg_info = _ACTIONS.setdefault(package, {})
    pkg_info[action] = ActionT(name=action, cls=cls, kegex=re.compile(cls.KEGEX))  # type: ignore[arg-type]
    return cls


def get(package: str, action: str) -> Callable[..., Any]:
    """Import and return a given action.

    Args:
        package: The name of the package
        action: The name of the action

    Returns:
        The action's registered class
    """
    _import(package, action)
    return _ACTIONS[package][action].cls


def get_factory(package: str) -> Callable[..., Any]:
    """Create a ``get()`` function for one package.

    Args:
        package: The name of the package

    Returns:
        The action's registered class
    """
    return functools.partial(get, package)


def kegex(package: str, action: str) -> tuple[str, str, str]:
    """Return a tuple of name, class, ``kegex`` for an action.

    Args:
        package: The name of the package
        action: The name of the action

    Returns:
        The name, class and kegex for an action
    """
    _import(package, action)
    return _ACTIONS[package][action]


def kegexes(package: str) -> Generator[tuple[str, str, str]]:
    """Return a tuple of tuples, name, ``kegex`` for all actions.

    Args:
        package: The name of the package

    Returns:
        A generator for all ``kegexes``
    """
    _import_all(package)
    return (kegex(package, name) for name in names(package))


def kegexes_factory(package: str) -> Callable[..., Any]:
    """Create a ``kegexes()`` function for all packages.

    Args:
        package: The name of the package

    Returns:
        A ``kegexes()`` method for the package
    """
    return functools.partial(kegexes, package)


def names(package: str) -> list[str]:
    """List all actions in one package.

    Args:
        package: The name of the package

    Returns:
        All packages
    """
    _import_all(package)
    return sorted(_ACTIONS[package])


def names_factory(package: str) -> Callable[..., Any]:
    """Create a ``names()`` function for one package.

    Args:
        package: The name of the package

    Returns:
        a ``names()`` method for the package
    """
    return functools.partial(names, package)


def run_interactive(package: str, action: str, *args: Any, **_kwargs: dict[str, Any]) -> Any:
    """Call the given action's ``run()`` method.

    Args:
        package: The name of the package
        action: The name of the action
        *args: The arguments passed to the action's run method
        **_kwargs: The keyword arguments passed to the action's run
            method

    Returns:
        The outcome of running the action's run method
    """
    action_cls = get(package, action)
    app, interaction = args
    app_action = action_cls(app.args)
    supports_interactive = hasattr(app_action, "run")
    if not supports_interactive:
        logger.error("Subcommand '%s' does not support mode interactive", action)
    run_action = app_action.run if supports_interactive else app_action.no_interactive_mode

    # Allow tracebacks to bring down the UI, used in tests
    if os.getenv("ANSIBLE_NAVIGATOR_ALLOW_UI_TRACEBACK") == "true":
        return run_action(app=app, interaction=interaction)

    # Capture and show a UI notification
    try:
        return run_action(app=app, interaction=interaction)
    except Exception:
        logger.critical("Subcommand '%s' encountered a fatal error.", action)
        logger.exception("Logging an uncaught exception")
        warn_msg = [f"Unexpected errors were encountered while running '{action}'."]
        warn_msg.append("Please log an issue with the log file contents.")
        warning = error_notification(warn_msg)
        interaction.ui.show_form(warning)
        return None


def run_interactive_factory(package: str) -> Callable[..., Any]:
    """Create a ``run_interactive()`` function for one package.

    Args:
        package: The name of the package

    Returns:
        A partial ``run_interactive()`` method for the package
    """
    return functools.partial(run_interactive, package)


def run_stdout(package: str, action: str, *args: Any, **_kwargs: dict[str, Any]) -> RunStdoutReturn:
    """Call the given action's ``run_stdout`` method.

    Args:
        package: The name of the package
        action: The name of the action
        *args: The arguments passed to the action's run_stdout method
        **_kwargs: The keyword arguments passed to the action's
            run_stdout method

    Returns:
        The outcome of running the action's ``run_stdout()`` method
    """
    # Refers to the action's run_stdout in the first line, not this function
    action_cls = get(package, action)
    args = args[0]
    return action_cls(args).run_stdout()


def run_stdout_factory(package: str) -> Callable[..., Any]:
    """Create a ``run_stdout()`` function for one package.

    Args:
        package: The name of the package

    Returns:
        A partial ``run_stdout()`` method for the package
    """
    return functools.partial(run_stdout, package)



"""Form definition and form rendering."""

from __future__ import annotations

import curses

from curses import ascii as curses_ascii
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Any

from .curses_defs import CursesLine
from .curses_defs import CursesLinePart
from .curses_defs import CursesLines
from .curses_window import CursesWindow
from .curses_window import Window
from .field_button import FieldButton
from .field_checks import FieldChecks
from .field_curses_information import FieldCursesInformation
from .field_information import FieldInformation
from .field_radio import FieldRadio
from .field_text import FieldText
from .field_working import FieldWorking
from .form_defs import FormType
from .form_handler_text import FormHandlerText
from .sentinels import unknown
from .validators import FormValidators


if TYPE_CHECKING:
    from ansible_navigator.ui_framework.ui_config import UIConfig


@dataclass
class Form:
    """Simple abstraction to hold the fields of the form and a convenience method to present it."""

    type_: FormType
    cancelled: bool = False
    fields: list[Any] = field(default_factory=list)
    submitted: bool = False
    title: str = ""
    title_color: int = 0

    _dict: dict[Any, Any] = field(default_factory=dict)

    def present(self, screen: Window, ui_config: UIConfig) -> Form:
        """Present the form the to user and return the results.

        Returns:
            Results from the form

        Args:
            screen: A curses window
            ui_config: The current user interface configuration
        """
        if self.type_ is FormType.FORM:
            self.fields.append(
                FieldButton(
                    name="submit",
                    text="Submit",
                    validator=FormValidators.all_true,
                    color=10,
                ),
            )
            self.fields.append(FieldButton(name="cancel", text="Cancel", color=9))
        elif self.type_ is FormType.NOTIFICATION:
            self.fields.append(
                FieldButton(
                    name="submit",
                    text=" Ok ",
                    validator=FormValidators.no_validation,
                    color=10,
                ),
            )
        elif self.type_ is FormType.WORKING:
            pass  # no additional fields needed for WORKING form type

        FormPresenter(form=self, screen=screen, ui_config=ui_config).present()
        try:
            self.submitted = next(field for field in self.fields if field.name == "submit").pressed
        except StopIteration:
            self.submitted = False
        try:
            self.cancelled = next(field for field in self.fields if field.name == "cancel").pressed
        except StopIteration:
            self.cancelled = False
        return self


# the maximum form size
MAX_FORM_HEIGHT = 1024
MAX_FORM_WIDTH = 1024
LEFT_PAD_RATIO = 2 / 5
TOP_PAD_RATIO = 2 / 5
BUTTON_SPACE = 10


class FormPresenter(CursesWindow):
    """Present the form to the user."""

    # pylint: disable=too-many-instance-attributes
    def __init__(self, form: Form, screen: Window, ui_config: UIConfig) -> None:
        """Initialize the form presenter.

        Args:
            form: The form to present to the user
            screen: A curses window
            ui_config: The current user interface configuration
        """
        super().__init__(ui_config=ui_config)
        self._form = form
        self._screen = screen
        self._line_number: int = 0
        self._prompt_end: int = 0
        self._input_start: int = 0
        self._form_width: int = 0
        self._form_height: int = 0
        self._pad_left: int = 0
        self._pad_top: int = 0
        self._separator = ": "

    @property
    def _field_win_start(self) -> int:
        """The window start of a field.

        Returns:
            The window start of a field
        """
        return self._input_start + self._pad_left

    @property
    def _field_win_width(self) -> int:
        """The window width of a field.

        Returns:
            The window width of a field
        """
        return self._screen_width - self._field_win_start

    def _calculate_field_widths(self, fields: list[Any]) -> list[int]:
        """Calculate the widths needed for each form field.

        Args:
            fields: The list of form fields

        Returns:
            A list of widths for each field
        """
        widths = []
        for form_field in fields:
            if hasattr(form_field, "value") and form_field.value is not unknown:
                widths.append(len(str(form_field.value)) + self._input_start)
            if hasattr(form_field, "options"):
                widths.extend(len(option.text) + self._input_start for option in form_field.options)
            if hasattr(form_field, "information"):
                widths.append(max(len(info) for info in form_field.information))
            if hasattr(form_field, "messages"):
                widths.append(max(len(msg) for msg in form_field.messages))
            widths.append(len(form_field.validator(hint=True)) + self._input_start)
        return widths

    def _calculate_form_height(self, fields: list[Any]) -> int:
        """Calculate the height of the form from fields.

        Args:
            fields: The list of form fields

        Returns:
            The total height of the form
        """
        height = 2  # title + horizontal line
        for form_field in fields:
            if isinstance(form_field, FieldCursesInformation | FieldInformation):
                height += len(form_field.information)
            if isinstance(form_field, (FieldWorking)):
                height += len(form_field.messages)
            elif isinstance(form_field, FieldText):
                height += 1
            elif isinstance(form_field, FieldChecks | FieldRadio):
                height += len(form_field.options)
        height += 2  # horizontal line + buttons
        return height

    def _dimensions(self) -> None:
        """Calculate the dimensions of the form."""
        self._prompt_end = max(len(form_field.full_prompt) for form_field in self._form.fields)
        self._input_start = self._prompt_end + len(self._separator)

        widths = self._calculate_field_widths(self._form.fields)

        if self._form.type_ is FormType.FORM:
            self._form_width = max(widths) + BUTTON_SPACE
        elif self._form.type_ in (FormType.NOTIFICATION, FormType.WORKING):
            self._form_width = max(widths)

        self._form_height = self._calculate_form_height(self._form.fields)

        self._pad_top = max(int((self._screen_height - self._form_height) * TOP_PAD_RATIO), 0)
        self._pad_left = max(int((self._screen_width - self._form_width) * LEFT_PAD_RATIO), 0)

    def _generate_body_field_lines(
        self,
        form_field: Any,
    ) -> list[tuple[int, CursesLine]]:
        """Generate the lines for a single body field.

        Args:
            form_field: The form field to generate lines for

        Returns:
            A list of (line_number, CursesLine) tuples for this field
        """
        lines: list[tuple[int, CursesLine]] = []

        if isinstance(form_field, (FieldCursesInformation)):
            for line in form_field.information:
                lines.append((self._line_number, line))
                self._line_number += 1

        elif isinstance(form_field, FieldInformation):
            information_lines = self._generate_information(form_field)
            for line in information_lines:  # pylint: disable=not-an-iterable
                lines.append((self._line_number, line))
                self._line_number += 1

        elif isinstance(form_field, FieldWorking):
            message_lines = self._generate_messages(form_field)
            for line in message_lines:  # pylint: disable=not-an-iterable
                lines.append((self._line_number, line))
                self._line_number += 1

        elif isinstance(form_field, FieldText):
            prompt = self._generate_prompt(form_field)
            line = CursesLine(tuple(prompt + [self._generate_field_text(form_field)]))
            lines.append((self._line_number, line))
            self._line_number += 1

        elif isinstance(form_field, FieldChecks | FieldRadio):
            prompt = self._generate_prompt(form_field)
            option_lines = self._generate_field_options(form_field)
            # although option_lines[0] is a CursesLine, only it's first line part is needed
            # because the prompt needs to be prepended to it
            first_option_line_part = option_lines[0][0]
            line = CursesLine(tuple(prompt + [first_option_line_part]))
            lines.append((self._line_number, line))
            self._line_number += 1

            for option_line in option_lines[1:]:
                lines.append((self._line_number, CursesLine(option_line)))
                self._line_number += 1

        error = self._generate_error(form_field)
        if error:
            lines.append((self._line_number, error))
            self._line_number += 1

        return lines

    def _generate_form(self) -> tuple[tuple[int, CursesLine], ...]:
        """Generate the form.

        Returns:
            A tuple of tuples containing the line numbers and the lines
        """
        lines = []
        lines.append((self._line_number, self._generate_title()))
        self._line_number += 1
        lines.append((self._line_number, self._generate_horizontal_line()))
        self._line_number += 1

        body_fields = [
            field for field in self._form.fields if field.name not in ["submit", "cancel"]
        ]
        for form_field in body_fields:
            lines.extend(self._generate_body_field_lines(form_field))

        lines.append((self._line_number, self._generate_horizontal_line()))
        self._line_number += 1
        lines.append((self._line_number, self._generate_buttons()))
        return tuple(lines)

    def _generate_buttons(self) -> CursesLine:
        """Generate the buttons.

        Returns:
            A CursesLine containing the buttons
        """
        line_parts = []
        far_right = self._form_width
        footer_fields = [field for field in self._form.fields if field.name in ["submit", "cancel"]]
        for form_field in reversed(footer_fields):
            string = f" {form_field.text} "  # room for []
            far_right -= len(string)
            window = curses.newwin(1, len(string), self._line_number, far_right + self._pad_left)
            window.keypad(True)
            form_field.win = window
            form_validity_state = [
                f.valid for f in self._form.fields if not isinstance(f, FieldButton)
            ]
            form_field.conditional_validation(form_validity_state)
            color = 8 if form_field.disabled is True else form_field.color
            line_part = CursesLinePart(far_right, string, color, 0)
            line_parts.append(line_part)
            far_right -= 1
        return CursesLine(tuple(line_parts))

    def _generate_error(self, form_field: FieldText) -> CursesLine | None:
        """Generate the error for a field.

        Args:
            form_field: The field to generate the error for

        Returns:
            A CursesLine containing the error
        """
        if form_field.current_error:
            line_part = CursesLinePart(self._input_start, form_field.current_error, 9, 0)
            return CursesLine((line_part,))
        return None

    def _generate_field_options(self, form_field: FieldChecks | FieldRadio) -> CursesLines:
        """Generate the options for a field.

        Args:
            form_field: The field to generate the options for

        Returns:
            A CursesLines containing the options
        """
        window = curses.newwin(
            len(form_field.options),
            self._field_win_width,
            self._line_number,
            self._field_win_start,
        )
        window.keypad(True)
        form_field.win = window
        lines = []
        for option in form_field.options:
            option_code = option.ansi_code(form_field)
            color = 8 if option.disabled else 0
            text = f"{option_code} {option.text!s}"
            line_part = CursesLinePart(self._input_start, text, color, 0)
            lines.append(CursesLine((line_part,)))
        return CursesLines(tuple(lines))

    def _generate_field_text(self, form_field: FieldText) -> CursesLinePart:
        """Generate the text for a field.

        Args:
            form_field: The field to generate the text for

        Returns:
            A CursesLinePart containing the text
        """
        window = curses.newwin(1, self._field_win_width, self._line_number, self._field_win_start)
        window.keypad(True)
        form_field.win = window

        if form_field.value is unknown:
            text = form_field.validator(hint=True)
            color = 8
        else:
            text = str(form_field.value)
            color = 0
        return CursesLinePart(self._input_start, text, color, 0)

    def _generate_horizontal_line(self) -> CursesLine:
        """Generate a horizontal line.

        Returns:
            A CursesLine containing the horizontal line
        """
        line_part = CursesLinePart(0, "\u2500" * self._form_width, 8, 0)
        return CursesLine((line_part,))

    @staticmethod
    def _generate_information(form_field: FieldInformation) -> CursesLines:
        """Generate an information field.

        Args:
            form_field: The field to generate the information for

        Returns:
            A CursesLines containing the information
        """
        lines = tuple(
            CursesLine((CursesLinePart(0, line, 0, 0),)) for line in form_field.information
        )
        return CursesLines(lines)

    @staticmethod
    def _generate_messages(form_field: FieldWorking) -> CursesLines:
        """Generate a messages field.

        Args:
            form_field: The field to generate the messages for

        Returns:
            A CursesLines containing the messages
        """
        lines = tuple(CursesLine((CursesLinePart(0, line, 0, 0),)) for line in form_field.messages)
        return CursesLines(lines)

    def _generate_prompt(
        self,
        form_field: FieldText | FieldChecks | FieldRadio,
    ) -> list[CursesLinePart]:
        """Generate the prompt for a field.

        Args:
            form_field: The field to generate the prompt for

        Returns:
            A list of CursesLinePart containing the prompt
        """
        prompt_start = self._prompt_end - len(form_field.full_prompt)
        color = 10 if form_field.valid is True else 0

        cl_prompt = CursesLinePart(prompt_start, form_field.prompt, color, 0)
        cl_default = CursesLinePart(
            prompt_start + len(form_field.prompt),
            str(form_field.formatted_default),
            4,
            0,
        )
        cl_separator = CursesLinePart(self._prompt_end, self._separator, color, 0)
        return [cl_prompt, cl_default, cl_separator]

    def _generate_title(self) -> CursesLine:
        """Generate the title for the form.

        Returns:
            A CursesLine containing the title
        """
        title = self._form.title
        if title.isupper() or not title[0].isupper():
            title = title.capitalize()
        line_part = CursesLinePart(0, title, self._form.title_color, 0)
        return CursesLine((line_part,))

    def _handle_form_input(
        self,
        idx: int,
        form_field: Any,
        win_response: tuple[Any, int],
    ) -> tuple[bool, int]:
        """Handle user input for a single form interaction cycle.

        Args:
            idx: The current field index
            form_field: The current form field
            win_response: The response and character from the window handler

        Returns:
            A tuple of (should_break, new_idx)
        """
        response, char = win_response

        if char == curses.KEY_RESIZE:
            self._screen.clear()
            self._screen.refresh()

        elif char == 112065:
            # non-blocking form
            return True, idx

        elif isinstance(form_field, FieldButton):
            if form_field.pressed:
                return True, idx
            idx += 1
        elif char == curses_ascii.TAB:
            form_field.conditional_validation(response)
            idx += 1
        else:
            form_field.validate(response)
            if form_field.valid is True:
                idx += 1

        return False, idx

    def present(self) -> Form:
        """Present the form to the user.

        Returns:
            Form to be shown to the user
        """
        self._screen.clear()
        self._screen.refresh()
        idx = 0
        pad = curses.newpad(MAX_FORM_HEIGHT, MAX_FORM_WIDTH)
        shared_input_line_cache: list[str] = []
        for form_field in self._form.fields:
            form_field.window_handler = form_field.window_handler(
                screen=self._screen,
                ui_config=self._ui_config,
            )
            if isinstance(form_field.window_handler, FormHandlerText):
                form_field.window_handler.input_line_cache = shared_input_line_cache

        while True:
            self._dimensions()
            self._line_number = self._pad_top

            pad.clear()
            for line in self._generate_form():
                self._add_line(pad, *line)
            pad.refresh(0, 0, 0, self._pad_left, self._screen_height - 1, self._screen_width - 1)

            idx = idx % len(self._form.fields)
            form_field = self._form.fields[idx]

            form_field.window_handler.win = form_field.win
            win_response = form_field.window_handler.handle(idx, self._form.fields)

            should_break, idx = self._handle_form_input(idx, form_field, win_response)
            if should_break:
                break

        return self._form


__all__ = ("FormType",)



"""Individual field check, the form field checks and radio check."""

from __future__ import annotations

import sys

from dataclasses import dataclass
from dataclasses import field
from functools import partial
from typing import TYPE_CHECKING
from typing import Any

from .form_handler_options import FormHandlerOptions
from .sentinels import Unknown
from .sentinels import unknown
from .validators import FieldValidators
from .validators import Validation


if TYPE_CHECKING:
    from collections.abc import Callable

    from ansible_navigator.ui_framework.curses_window import Window


@dataclass
class FieldChecks:
    # pylint: disable=too-many-instance-attributes
    """A form field containing checks."""

    prompt: str
    name: str
    current_error: str = ""
    valid: Unknown | bool = unknown
    options: list[Any] = field(default_factory=list)
    max_selected: int = sys.maxsize
    min_selected: int = 1
    window_handler = FormHandlerOptions
    win: Window | None = None

    @property
    def checked(self) -> tuple[bool, ...]:
        """Conveniently return just checked fields.

        Returns:
            Name of every checked field
        """
        return tuple(option.name for option in self.options if option.checked)

    @property
    def formatted_default(self) -> str:
        """Format the checked field prompt with an empty string.

        Returns:
            Empty string
        """
        return ""

    @property
    def full_prompt(self) -> str:
        """Format the checkbox prompt.

        Returns:
            Checkbox prompt
        """
        return self.prompt

    @property
    def validator(self) -> Callable[..., Any]:
        """Provide a validator based on form type.

        Returns:
            Validation of checked entries
        """
        return partial(
            FieldValidators.some_of_or_none,
            max_selected=self.max_selected,
            min_selected=self.min_selected,
        )

    def _validate(self, response: FieldChecks) -> Validation:
        """Validate this FieldChecks instance.

        Args:
            response: The form response from the user

        Returns:
            Validation of the response
        """
        validation = self.validator(choices=response.options)
        if validation.error_msg:
            self.valid = False
        else:
            self.valid = True
        return validation

    def validate(self, response: FieldChecks) -> None:
        """Validate this FieldChecks instance.

        Args:
            response: Instance to check and verify options are valid
        """
        if self.max_selected == sys.maxsize:
            self.max_selected = len(self.options)

        validation = self._validate(response)
        self.current_error = validation.error_msg

    def conditional_validation(self, response: FieldChecks) -> None:
        """Conditional validation for a field_checks instance.

        Args:
            response: Instance to check and verify options are valid
        """
        self._validate(response)
        self.current_error = ""



from __future__ import annotations

import functools

from typing import TYPE_CHECKING

from .reg import make_regset
from .rules import EndRule
from .rules import Entry
from .rules import MatchRule
from .rules import PatternRule
from .rules import WhileRule
from .state import State


if TYPE_CHECKING:
    from .fchainmap import FChainMap
    from .grammars import Grammar
    from .grammars import Grammars
    from .rules import Captures
    from .rules import CompiledRule
    from .rules import _Rule

INCLUDE_SELF = "$self"


class Compiler:
    def __init__(self, grammar: Grammar, grammars: Grammars) -> None:
        """Initialize the grammar compiler.

        Args:
            grammar: The grammar to compile, a text mate language file
            grammars: All previously compiled grammars
        """
        self._root_scope = grammar.scope_name
        self._grammars = grammars
        self._rule_to_grammar: dict[_Rule, Grammar] = {}
        self._c_rules: dict[_Rule, CompiledRule] = {}
        root = self._compile_root(grammar)
        self.root_state = State.root(Entry(root.name, root, ("", 0)))

    def _visit_rule(self, grammar: Grammar, rule: _Rule) -> _Rule:
        self._rule_to_grammar[rule] = grammar
        return rule

    @functools.cache  # noqa: B019
    def _include(
        self,
        grammar: Grammar,
        repository: FChainMap[str, _Rule],
        s: str,
    ) -> tuple[list[str], tuple[_Rule, ...]]:
        if s == INCLUDE_SELF:
            return self._patterns(grammar, grammar.patterns)
        if s == "$base":
            grammar = self._grammars.grammar_for_scope(self._root_scope)
            return self._include(grammar, grammar.repository, INCLUDE_SELF)
        if s.startswith("#"):
            return self._patterns(grammar, (repository[s[1:]],))
        if "#" not in s:
            grammar = self._grammars.grammar_for_scope(s)
            return self._include(grammar, grammar.repository, INCLUDE_SELF)
        scope, _, s = s.partition("#")
        grammar = self._grammars.grammar_for_scope(scope)
        return self._include(grammar, grammar.repository, f"#{s}")

    @functools.cache  # noqa: B019
    def _patterns(
        self,
        grammar: Grammar,
        rules: tuple[_Rule, ...],
    ) -> tuple[list[str], tuple[_Rule, ...]]:
        ret_regs = []
        ret_rules: list[_Rule] = []
        for rule in rules:
            if rule.include is not None:
                tmp_regs, tmp_rules = self._include(grammar, rule.repository, rule.include)
                ret_regs.extend(tmp_regs)
                ret_rules.extend(tmp_rules)
            elif rule.match is None and rule.begin is None and rule.patterns:
                tmp_regs, tmp_rules = self._patterns(grammar, rule.patterns)
                ret_regs.extend(tmp_regs)
                ret_rules.extend(tmp_rules)
            elif rule.match is not None:
                ret_regs.append(rule.match)
                ret_rules.append(self._visit_rule(grammar, rule))
            elif rule.begin is not None:
                ret_regs.append(rule.begin)
                ret_rules.append(self._visit_rule(grammar, rule))
            else:
                msg = f"unreachable {rule}"
                raise AssertionError(msg)
        return ret_regs, tuple(ret_rules)

    def _captures_ref(self, grammar: Grammar, captures: Captures) -> Captures:
        return tuple((n, self._visit_rule(grammar, r)) for n, r in captures)

    def _compile_root(self, grammar: Grammar) -> PatternRule:
        regs, rules = self._patterns(grammar, grammar.patterns)
        return PatternRule((grammar.scope_name,), make_regset(*regs), rules)

    def _compile_rule(self, grammar: Grammar, rule: _Rule) -> CompiledRule:
        assert rule.include is None, rule  # noqa:S101
        if rule.match is not None:
            captures_ref = self._captures_ref(grammar, rule.captures)
            return MatchRule(rule.name, captures_ref)
        if rule.begin is not None and rule.end is not None:
            regs, rules = self._patterns(grammar, rule.patterns)
            return EndRule(
                rule.name,
                rule.content_name,
                self._captures_ref(grammar, rule.begin_captures),
                self._captures_ref(grammar, rule.end_captures),
                rule.end,
                make_regset(*regs),
                rules,
            )
        if rule.begin is not None and rule.while_ is not None:
            regs, rules = self._patterns(grammar, rule.patterns)
            return WhileRule(
                rule.name,
                rule.content_name,
                self._captures_ref(grammar, rule.begin_captures),
                self._captures_ref(grammar, rule.while_captures),
                rule.while_,
                make_regset(*regs),
                rules,
            )
        regs, rules = self._patterns(grammar, rule.patterns)
        return PatternRule(rule.name, make_regset(*regs), rules)

    def compile_rule(self, rule: _Rule) -> CompiledRule:
        try:
            return self._c_rules[rule]
        except KeyError:
            pass

        grammar = self._rule_to_grammar[rule]
        ret = self._c_rules[rule] = self._compile_rule(grammar, rule)
        return ret



"""The v1 to v2 settings file migration."""

from __future__ import annotations

from typing import Any

from ansible_navigator.utils.dot_paths import MergeBehaviors
from ansible_navigator.utils.dot_paths import check_path
from ansible_navigator.utils.dot_paths import get_with_path
from ansible_navigator.utils.dot_paths import move_to_path
from ansible_navigator.utils.dot_paths import place_at_path

from .definitions import MigrationStep
from .definitions import MigrationType
from .settings_file import SettingsFile


class V1V2SettingsFile(SettingsFile):
    """The v1 to v2 settings file migration.

    The ordering of the functions in this class is important.
    It ensures that the migrations are run in the correct order.
    """

    name = "Version 1 to Version 2 settings file format migration"
    migration_type: MigrationType = MigrationType.SETTINGS_FILE

    def __init__(self) -> None:
        """Initialize the v1 to v2 settings file migration."""
        super().__init__()
        self.content: dict[Any, Any] = {}
        self._backup_suffix = ".v1"

    @MigrationStep.register(MigrationStep(name="config path"))
    def config_path(self) -> bool:
        """Migrate the config path entry.

        Returns:
            Whether the migration is needed
        """
        old_path = "ansible-navigator.ansible.config"
        new_path = "ansible-navigator.ansible.config.path"

        exists = check_path(self.content, old_path)
        if not exists:
            return False

        value = get_with_path(self.content, old_path)
        if not isinstance(value, str):
            return False

        if self.check:
            return True

        self.content = move_to_path(
            behaviors=(),
            content=self.content,
            new_path=new_path,
            old_path=old_path,
        )
        return False

    @MigrationStep.register(MigrationStep(name="documentation"))
    def documentation(self) -> bool:
        """Migrate the documentation entry.

        Returns:
            Whether the migration is needed
        """
        old_path = "ansible-navigator.documentation"
        new_path = "ansible-navigator.ansible.doc"

        exists = check_path(self.content, old_path)
        if self.check or not exists:
            return exists

        self.content = move_to_path(
            behaviors=(),
            content=self.content,
            new_path=new_path,
            old_path=old_path,
        )
        return False

    @MigrationStep.register(MigrationStep(name="playbook path"))
    def playbook_path(self) -> bool:
        """Migrate the playbook path entry.

        Returns:
            Whether the migration is needed
        """
        old_path = "ansible-navigator.ansible.playbook"
        new_path = "ansible-navigator.ansible.playbook.path"

        exists = check_path(self.content, old_path)
        if not exists:
            return False

        value = get_with_path(self.content, old_path)
        if not isinstance(value, str):
            return False

        if self.check:
            return True

        self.content = move_to_path(
            behaviors=(),
            content=self.content,
            new_path=new_path,
            old_path=old_path,
        )
        return False

    @MigrationStep.register(MigrationStep(name="help builder"))
    def help_builder(self) -> bool:
        """Migrate the help-builder entry.

        Returns:
            Whether the migration is needed
        """
        old_path = "ansible-navigator.help-builder"
        new_path = "ansible-navigator.ansible-builder.help"

        exists = check_path(self.content, old_path)
        if self.check or not exists:
            return exists

        self.content = move_to_path(
            behaviors=(),
            content=self.content,
            new_path=new_path,
            old_path=old_path,
        )
        return False

    @MigrationStep.register(MigrationStep(name="help config"))
    def help_config(self) -> bool:
        """Migrate the help-config entry.

        Returns:
            Whether the migration is needed
        """
        old_path = "ansible-navigator.help-config"
        new_path = "ansible-navigator.ansible.config.help"

        exists = check_path(self.content, old_path)
        if self.check or not exists:
            return exists

        self.content = move_to_path(
            behaviors=(),
            content=self.content,
            new_path=new_path,
            old_path=old_path,
        )
        return False

    @MigrationStep.register(MigrationStep(name="help doc"))
    def help_doc(self) -> bool:
        """Migrate the help-doc entry.

        Returns:
            Whether the migration is needed
        """
        old_path = "ansible-navigator.help-doc"
        new_path = "ansible-navigator.ansible.doc.help"

        exists = check_path(self.content, old_path)
        if self.check or not exists:
            return exists

        self.content = move_to_path(
            behaviors=(),
            content=self.content,
            new_path=new_path,
            old_path=old_path,
        )
        return False

    @MigrationStep.register(MigrationStep(name="help inventory"))
    def help_inventory(self) -> bool:
        """Migrate the help-inventory entry.

        Returns:
            Whether the migration is needed
        """
        old_path = "ansible-navigator.help-inventory"
        new_path = "ansible-navigator.ansible.inventory.help"

        exists = check_path(self.content, old_path)
        if self.check or not exists:
            return exists

        self.content = move_to_path(
            behaviors=(),
            content=self.content,
            new_path=new_path,
            old_path=old_path,
        )
        return False

    @MigrationStep.register(MigrationStep(name="help playbook"))
    def help_playbook(self) -> bool:
        """Migrate the help-playbook entry.

        Returns:
            Whether the migration is needed
        """
        old_path = "ansible-navigator.help-playbook"
        new_path = "ansible-navigator.ansible.playbook.help"

        exists = check_path(self.content, old_path)
        if self.check or not exists:
            return exists

        self.content = move_to_path(
            behaviors=(),
            content=self.content,
            new_path=new_path,
            old_path=old_path,
        )
        return False

    @MigrationStep.register(MigrationStep(name="inventory paths"))
    def inventories(self) -> bool:
        """Migrate the inventory paths.

        Returns:
            Whether the migration is needed
        """
        old_path = "ansible-navigator.ansible.inventories"
        new_path = "ansible-navigator.ansible.inventory.entries"

        exists = check_path(self.content, old_path)
        if self.check or not exists:
            return exists

        self.content = move_to_path(
            behaviors=(
                MergeBehaviors.LIST_APPEND,
                MergeBehaviors.LIST_SORT,
                MergeBehaviors.LIST_UNIQUE,
            ),
            content=self.content,
            old_path=old_path,
            new_path=new_path,
        )
        return False

    @MigrationStep.register(MigrationStep(name="playbook artifact timestamp"))
    def playbook_artifact(self) -> bool:
        """Migrate the playbook artifact entry.

        Returns:
            Whether the migration is needed
        """
        path = "ansible-navigator.playbook-artifact.save-as"

        entry_exists = check_path(self.content, path)
        if not entry_exists:
            return False

        value = get_with_path(self.content, path)

        if not isinstance(value, str):
            return False

        if self.check:
            return "ts_utc" in value

        new_value = value.replace("ts_utc", "time_stamp")

        self.content = place_at_path(
            behaviors=(),
            content=self.content,
            path=path,
            value=new_value,
        )
        return False

    @MigrationStep.register(MigrationStep(name="pull-policy"))
    def pull_policy(self) -> bool:
        """Migrate the pull-policy entry.

        Returns:
            Whether the migration is needed
        """
        old_path = "ansible-navigator.execution-environment.pull-policy"
        new_path = "ansible-navigator.execution-environment.pull.policy"

        exists = check_path(self.content, old_path)
        if self.check or not exists:
            return exists

        self.content = move_to_path(
            behaviors=(),
            content=self.content,
            new_path=new_path,
            old_path=old_path,
        )
        return False

    @MigrationStep.register(MigrationStep(name="volume mount labels"))
    def volume_mount_labels(self) -> bool:
        """Migrate the volume mount labels.

        Returns:
            Whether the migration is needed
        """
        path = "ansible-navigator.execution-environment.volume-mounts"

        entry_exists = check_path(self.content, path)
        if not entry_exists:
            return False

        value = get_with_path(self.content, path)

        if not isinstance(value, list):
            return False

        if not all(isinstance(item, dict) for item in value):
            return False

        if self.check:
            return any("label" in item for item in value)

        for item in value:
            if "label" in item:
                item["options"] = item.pop("label")

        self.content = place_at_path(
            behaviors=(MergeBehaviors.LIST_LIST_REPLACE,),
            content=self.content,
            path=path,
            value=value,
        )

        return False



"""Constants for use with the user interface."""

from __future__ import annotations

import curses

from enum import IntEnum


class Color(IntEnum):
    """Constants for the 16 basic ANSI colors.

    Attributes:
        BLACK: Color index 0.
        RED: Color index 1.
        GREEN: Color index 2.
        YELLOW: Color index 3.
        BLUE: Color index 4.
        MAGENTA: Color index 5.
        CYAN: Color index 6.
        WHITE: Color index 7.
        GREY: Color index 8.
        BRIGHT_RED: Color index 9.
        BRIGHT_GREEN: Color index 10.
        BRIGHT_YELLOW: Color index 11.
        BRIGHT_BLUE: Color index 12.
        BRIGHT_MAGENTA: Color index 13.
        BRIGHT_CYAN: Color index 14.
        BRIGHT_WHITE: Color index 15.
    """

    BLACK = curses.COLOR_BLACK
    RED = curses.COLOR_RED
    GREEN = curses.COLOR_GREEN
    YELLOW = curses.COLOR_YELLOW
    BLUE = curses.COLOR_BLUE
    MAGENTA = curses.COLOR_MAGENTA
    CYAN = curses.COLOR_CYAN
    WHITE = curses.COLOR_WHITE
    GREY = BLACK + 8
    BRIGHT_RED = RED + 8
    BRIGHT_GREEN = GREEN + 8
    BRIGHT_YELLOW = YELLOW + 8
    BRIGHT_BLUE = BLUE + 8
    BRIGHT_MAGENTA = MAGENTA + 8
    BRIGHT_CYAN = CYAN + 8
    BRIGHT_WHITE = WHITE + 8


class Decoration(IntEnum):
    """Constants for text decoration.

    These are rarely used within the interface.

    Attributes:
        BOLD: Bold text attribute.
        ITALIC: Italic text attribute.
        NORMAL: Normal text attribute.
        UNDERLINE: Underline text attribute.
    """

    BOLD = curses.A_BOLD
    # Fix for missing curses.A_ITALIC on macOS
    ITALIC = getattr(curses, "A_ITALIC", curses.A_BOLD)
    NORMAL = curses.A_NORMAL
    UNDERLINE = curses.A_UNDERLINE

    @classmethod
    def get_best(cls, name: str | None) -> int:
        """Return the default value for a missing value.

        Args:
            name: The name of the value

        Returns:
            The match or default value
        """
        if isinstance(name, str):
            try:
                return cls[name.upper()]
            except KeyError:
                return cls.NORMAL
        return cls.NORMAL



"""Ansible-navigator version information."""

try:
    from ._version import version as __version__
except ImportError:  # pragma: no branch
    try:
        from importlib.metadata import version

        __version__ = version("ansible-navigator")
    except Exception:  # noqa: BLE001
        # this is the fallback SemVer version picked by setuptools_scm when tag
        # information is not available.
        __version__ = "0.1.dev1"

__all__ = ("__version__",)



"""A field with a 'working on it message'."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any

from .form_handler_working import FormHandlerWorking
from .sentinels import Unknown
from .sentinels import unknown
from .validators import FieldValidators


if TYPE_CHECKING:
    from collections.abc import Callable

    from .curses_window import Window


@dataclass
class FieldWorking:
    """A text input field."""

    name: str
    messages: list[str]
    current_error: str = ""
    window_handler = FormHandlerWorking
    valid: bool | Unknown = unknown
    validator: Callable[..., Any] = FieldValidators.null
    win: Window | None = None

    @property
    def full_prompt(self) -> str:
        """Return the max width information.

        This is needed because windows width and : placement
        is based on the largest 'prompt'.

        Returns:
            Max width information message
        """
        return max(self.messages)

    def validate(self, response: str) -> None:
        """No validation required for working field.

        Args:
            response: Field response
        """
        self.valid = True

    def conditional_validation(self, response: str) -> None:
        """No conditional validation.

        Args:
            response: Field response
        """
        self.validate(response)



"""Pre-defined objects used by actions."""

from __future__ import annotations

from typing import NamedTuple


class RunReturn(NamedTuple):
    """The base return object for any action's run method."""

    #: A message to display immediately before exiting
    message: str
    #: The return code for the for the process
    return_code: int


class RunInteractiveReturn(RunReturn):
    """The return object for an action's run interactive method."""


class RunStdoutReturn(RunReturn):
    """The return object for an action's run stdout method."""


ActionReturn = RunReturn | RunInteractiveReturn | RunStdoutReturn



"""Herein lies the ability for ansible-runner to run the ansible-config command."""

from __future__ import annotations

import warnings


# Remove this catch-all once newer ansible-runner is released
# https://github.com/ansible/ansible-runner/issues/1223
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from ansible_runner import get_ansible_config

from .base import Base


class AnsibleConfig(Base):
    """Abstraction for ansible-config command-line."""

    def fetch_ansible_config(
        self,
        action: str,
        config_file: str | None = None,
        only_changed: bool | None = None,
    ) -> tuple[str, str]:
        """Run ansible-config command and get the configuration related details.

        Args:
            action: The configuration fetch action to perform. Valid
                values are one of ``list``, ``dump``, ``view``. The
                ``list`` action will fetch all the config options along
                with config description, ``dump`` action will fetch all
                the active config and ``view`` action will return the
                active configuration file view.
            config_file: Path to configuration file, defaults to first
                file found in precedence. Defaults to ``None``.
            only_changed: The boolean value when set to ``True`` returns
                only the configurations that have changed from the
                default. This parameter is applicable only when
                ``action`` is set to ``dump``. Defaults to `None`.

        Returns:
            A tuple of response and error string (if any)
        """
        return get_ansible_config(
            action,
            config_file=config_file,
            only_changed=only_changed,
            **self._runner_args,
        )



"""Build the argument parser."""

from __future__ import annotations

import argparse

from argparse import SUPPRESS
from argparse import ArgumentParser
from argparse import HelpFormatter
from argparse import _SubParsersAction
from typing import Any

from .definitions import ApplicationConfiguration
from .definitions import Constants as C
from .definitions import SettingsEntry


class Parser:
    """Build the args."""

    def __init__(self, config: ApplicationConfiguration) -> None:
        """Initialize the command line interface parameter parser.

        Args:
            config: The current settings for the application
        """
        self._config = config
        self._base_parser = ArgumentParser(add_help=False)
        self._configure_base()
        self.parser = ArgumentParser(
            parents=[self._base_parser],
            formatter_class=CustomHelpFormatter,
            add_help=False,
        )
        self._subparsers = self._add_subcommand_holder()
        self._configure_subparsers()

    @staticmethod
    def generate_argument(entry: SettingsEntry) -> tuple[Any, Any, dict[str, Any]]:
        """Generate an argparse argument.

        Args:
            entry: Single settings entry

        Returns:
            Long and short cli parameters, and dictionary of parsed
            arguments

        Raises:
            NotImplementedError: if there is an error
        """
        kwargs = {}
        help_strings = [entry.short_description]
        if entry.choices:
            lower_choices = (str(choice).lower() for choice in entry.choices)
            help_strings.append(f"({'|'.join(lower_choices)})")

        has_default = entry.value.default is not C.NOT_SET
        if entry.cli_parameters is None:
            raise NotImplementedError
        not_store = entry.cli_parameters.action not in ("store_true", "store_false")
        if has_default and not_store:
            help_strings.append(f"(default: {str(entry.value.default).lower()})")
        kwargs["help"] = " ".join(help_strings)

        kwargs["default"] = SUPPRESS

        if entry.cli_parameters.positional:
            long = None
            if entry.cli_parameters.nargs is None:
                kwargs["nargs"] = "?"
            else:
                kwargs["nargs"] = entry.cli_parameters.nargs
        else:
            long = entry.cli_parameters.long_override or f"--{entry.name_dashed}"
            kwargs["dest"] = entry.name
            if entry.cli_parameters.nargs is not None:
                kwargs["nargs"] = entry.cli_parameters.nargs
            if entry.cli_parameters.const is not None:
                kwargs["const"] = str(entry.cli_parameters.const)

        if entry.cli_parameters.action is not None:
            kwargs["action"] = entry.cli_parameters.action

        return entry.cli_parameters.short, long, kwargs

    def _add_parser(self, group: argparse._ArgumentGroup, entry: SettingsEntry) -> None:
        """Add a parser to the subparsers.

        Args:
            group: The group to add the parser to
            entry: The entry to add
        """
        if entry.cli_parameters:
            short, long, kwargs = self.generate_argument(entry)
            if not all((short, long)):
                group.add_argument(entry.name, **kwargs)
            else:
                group.add_argument(short, long, **kwargs)

    def _add_subcommand_holder(self) -> _SubParsersAction[Any]:
        """Add the subparsers holder.

        Raises:
            ValueError: if zero or more than one subcommand is found

        Returns:
            The subparsers action
        """
        subcommand_value = [
            entry for entry in self._config.entries if entry.subcommand_value is True
        ]
        if len(subcommand_value) == 0:
            msg = "No entry with subparser value defined"
            raise ValueError(msg)
        if len(subcommand_value) > 1:
            msg = "Multiple entries with subparser value defined"
            raise ValueError(msg)
        entry = subcommand_value[0]
        return self.parser.add_subparsers(
            title=entry.short_description,
            dest=entry.name,
            metavar="{subcommand} --help",
        )

    def _configure_base(self) -> None:
        """Configure the base parser."""
        group = self._base_parser.add_argument_group("Options (global)")
        group.add_argument(
            "-h",
            "--help",
            action="help",
            default=SUPPRESS,
            help="Show this help message and exit",
        )

        if isinstance(self._config.application_version, C):
            version = self._config.application_version.value
        else:
            version = self._config.application_version
        group.add_argument(
            "--version",
            action="version",
            help="Show the application version and exit",
            version="%(prog)s " + version,
        )

        for entry in self._config.entries:
            if entry.subcommands is C.ALL:
                self._add_parser(group=group, entry=entry)

    def _configure_subparsers(self) -> None:
        """Configure the subparsers."""
        for subcommand in self._config.subcommands:
            parser = self._subparsers.add_parser(
                subcommand.name,
                epilog=subcommand.epilog,
                help=subcommand.description,
                description=f"{subcommand.name}: {subcommand.description}",
                parents=[self._base_parser],
                formatter_class=CustomHelpFormatter,
                add_help=False,
            )
            group = parser.add_argument_group(f"Options ({subcommand.name} subcommand)")
            for entry in self._config.entries:
                if isinstance(entry.subcommands, list) and subcommand.name in entry.subcommands:
                    self._add_parser(group=group, entry=entry)


class CustomHelpFormatter(HelpFormatter):
    """A custom help formatter."""

    def __init__(self, prog: str) -> None:
        """Initialize the help formatter.

        Args:
            prog: The program name
        """
        long_string = "--rac  --ansible-runner-rotate-artifacts-count"
        # 3 here accounts for the spaces in the ljust(6) below
        HelpFormatter.__init__(
            self,
            prog=prog,
            indent_increment=1,
            max_help_position=len(long_string) + 3,
        )

    def _format_action_invocation(self, action: argparse.Action) -> str:
        """Format the action invocation.

        Args:
            action: The action to format

        Raises:
            ValueError: If more than 2 options are given

        Returns:
            The formatted action invocation
        """
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            (metavar,) = self._metavar_formatter(action, default)(1)
            return metavar

        if len(action.option_strings) == 1:
            return action.option_strings[0]

        if len(action.option_strings) == 2:
            # Account for a --1234 --long-option-name
            msg = f"{action.option_strings[0].ljust(6)} {action.option_strings[1]}"
            return msg
        msg = "Too many option strings"
        raise ValueError(msg)

    def _format_usage(self, usage: Any, actions: Any, groups: Any, prefix: str | None) -> str:
        """Format the usage.

        Args:
            usage: The usage
            actions: The actions
            groups: The groups
            prefix: The prefix

        Returns:
            The formatted usage
        """
        prefix = "Usage:" if prefix is None else ""
        options = "[options]" if actions else ""
        return " ".join(p for p in (prefix, self._prog, options) if p != "") + "\n\n"



"""``:sample_notification`` command implementation."""

from ansible_navigator.action_base import ActionBase
from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import dict_to_form
from ansible_navigator.ui_framework import form_to_dict
from ansible_navigator.utils.serialize import yaml

from . import _actions as actions


# cspell:disable
FORM = """
form:
  title:  BLOCKING NOTIFICATION
  title_color: 1
  fields:
    - name: info
      information:
      - Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut
      - labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco
      - laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in
      - velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non,
      - sunt in culpa qui officia deserunt mollit anim id est laborum.
      type: information
  type: notification
"""
# cspell:enable


@actions.register
class Action(ActionBase):
    """``:sample_notification`` command implementation."""

    KEGEX = r"^sample_notification$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:sample_notification`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="sample_form")

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction:
        """Execute the ``:sample_notification`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction`
        """
        self._logger.debug("sample notification requested")
        self._prepare_to_run(app, interaction)

        form_data = yaml.safe_load(FORM)
        form = dict_to_form(form_data["form"])
        interaction.ui.show_form(form)
        as_dict = form_to_dict(form)
        self._logger.debug("form response: %s", as_dict)

        while True:
            self._calling_app.update()
            next_interaction: Interaction = interaction.ui.show(obj=as_dict)
            if next_interaction.name != "refresh":
                break

        self._prepare_to_exit(interaction)
        return next_interaction



"""Individual check and the form field checks for radios."""

from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from functools import partial
from typing import TYPE_CHECKING
from typing import Any

from .form_handler_options import FormHandlerOptions
from .sentinels import Unknown
from .sentinels import unknown
from .validators import FieldValidators
from .validators import Validation


if TYPE_CHECKING:
    from collections.abc import Callable

    from ansible_navigator.ui_framework.curses_window import Window


@dataclass
class FieldRadio:
    """A form field containing radios."""

    prompt: str
    name: str
    current_error: str = ""
    valid: Unknown | bool = unknown
    options: list[Any] = field(default_factory=list)
    window_handler = FormHandlerOptions
    win: Window | None = None

    @property
    def checked(self) -> tuple[str, ...]:
        """Conveniently return just checked options.

        Returns:
            Checked options
        """
        return tuple(option.name for option in self.options if option.checked)

    @property
    def formatted_default(self) -> str:
        """Format the field prompt with an empty string.

        Returns:
            Empty string
        """
        return ""

    @property
    def full_prompt(self) -> str:
        """Format the prompt.

        Returns:
            Prompt
        """
        return self.prompt

    @property
    def validator(self) -> Callable[..., Validation]:
        """Provide a validator based on form type.

        Returns:
            Validation of checked entries

        Raises:
            TypeError: If there is a typing error.
        """
        result = partial(FieldValidators.some_of_or_none, max_selected=1, min_selected=1)
        if not isinstance(result, Validation):
            raise TypeError
        return result

    def _validate(self, response: FieldRadio) -> Validation:
        """Validate this FieldRadio instance.

        Args:
            response: The form response from the user

        Returns:
            Validation of the response
        """
        validation = self.validator(choices=response.options)
        if validation.error_msg:
            self.valid = False
        else:
            self.valid = True
        return validation

    def validate(self, response: FieldRadio) -> None:
        """Validate this FieldRadio instance.

        Args:
            response: Instance to check and verify options are valid
        """
        validation = self._validate(response)
        self.current_error = validation.error_msg

    def conditional_validation(self, response: FieldRadio) -> None:
        """Conditional validation for a FieldRadio instance.

        Args:
            response: Instance to check and verify options are valid
        """
        self._validate(response)
        self.current_error = ""



"""Herein lies the ability for ansible-runner to run the ansible-inventory command."""

from __future__ import annotations

from typing import Any

from ansible_runner import get_inventory

from .base import Base


class AnsibleInventory(Base):
    # pylint: disable=too-many-arguments
    """Abstraction for ansible-inventory command-line."""

    def fetch_inventory(
        self,
        action: str,
        inventories: list[Any],
        response_format: str | None = None,
        host: str | None = None,
        playbook_dir: str | None = None,
        vault_ids: str | None = None,
        vault_password_file: str | None = None,
    ) -> tuple[str, str]:
        """Run ansible-inventory command and get the inventory related details.

        Args:
            action: Valid values are one of ``graph``, ``host``,
                ``list``, ``graph`` create inventory graph, ``host``
                returns specific host info and works as inventory script
                and ``list`` output all hosts info and also works as
                inventory script.
            inventories: List of inventory host paths
            response_format: The output format for response. Valid
                values can be one of ``json``, ``yaml``, ``toml``. If
                ``action`` is ``graph`` only allowed value is ``json``.
            host: When ``action`` is set to ``host`` this parameter is
                used to get the host specific information.
            playbook_dir: This parameter is used to sets the relative
                path for the inventory
            vault_ids: The vault identity to use
            vault_password_file: The vault identity to use

        Returns:
            A tuple of response and error string (if any)
        """
        return get_inventory(
            action,
            inventories=inventories,
            response_format=response_format,
            host=host,
            playbook_dir=playbook_dir,
            vault_ids=vault_ids,
            vault_password_file=vault_password_file,
            **self._runner_args,
        )



"""An image introspector."""

from __future__ import annotations

import inspect
import json
import logging
import tempfile

from pathlib import Path
from typing import Any

from ansible_navigator.data import image_introspect
from ansible_navigator.runner import Command


logger = logging.getLogger(__name__)


def run(image_name: str, container_engine: str) -> tuple[dict[Any, Any], list[str], int]:
    """Run runner to collect image details.

    Args:
        image_name: The full image name
        container_engine: The container engine to use

    Returns:
        Output, errors and the return code
    """
    errors = []
    python_exec_path = "python3"

    with tempfile.TemporaryDirectory() as tmp_dir_name:
        introspect_source = inspect.getsource(image_introspect)
        file = Path(tmp_dir_name) / "image_introspect.py"
        file.write_text(introspect_source)

        _runner = Command(
            cmdline=[str(file)],
            container_engine=container_engine,
            container_volume_mounts=[f"{tmp_dir_name}:{tmp_dir_name}"],
            executable_cmd=python_exec_path,
            execution_environment_image=image_name,
            execution_environment=True,
            navigator_mode="interactive",
        )
        output, error, return_code = _runner.run()
    if error:
        errors.append(error)
    parse_errors, parsed = parse(output)
    errors.extend(parse_errors)
    return parsed, errors, return_code


def parse(output: str) -> tuple[list[str], dict[Any, Any]]:
    """Load and process the ``json`` output from the image introspection process.

    Args:
        output: The output from the image introspection process

    Returns:
        The parsed output
    """
    errors = []
    try:
        if not output.startswith("{"):
            _warnings, json_str = output.split("{", 1)
            json_str = "{" + json_str
        else:
            json_str = output
        parsed = json.loads(json_str)
    except (json.decoder.JSONDecodeError, ValueError) as exc:
        errors.append("Unable to extract introspection from stdout")
        errors.append(f"error json loading output: '{exc!s}'")
        errors.append(output)
        return errors, {}

    return errors, parsed
